#!/usr/bin/env python3
# coding: utf-8

# ###############################################################################
# This file is part of PanACOTA.                                                #
#                                                                               #
# Authors: Amandine Perrin                                                      #
# Copyright © 2018-2020 Institut Pasteur (Paris).                               #
# See the COPYRIGHT file for details.                                           #
#                                                                               #
# PanACOTA is a software providing tools for large scale bacterial comparative  #
# genomics. From a set of complete and/or draft genomes, you can:               #
#    -  Do a quality control of your strains, to eliminate poor quality         #
# genomes, which would not give any information for the comparative study       #
#    -  Uniformly annotate all genomes                                          #
#    -  Do a Pan-genome                                                         #
#    -  Do a Core or Persistent genome                                          #
#    -  Align all Core/Persistent families                                      #
#    -  Infer a phylogenetic tree from the Core/Persistent families             #
#                                                                               #
# PanACOTA is free software: you can redistribute it and/or modify it under the #
# terms of the Affero GNU General Public License as published by the Free       #
# Software Foundation, either version 3 of the License, or (at your option)     #
# any later version.                                                            #
#                                                                               #
# PanACOTA is distributed in the hope that it will be useful, but WITHOUT ANY   #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS     #
# FOR A PARTICULAR PURPOSE. See the Affero GNU General Public License           #
# for more details.                                                             #
#                                                                               #
# You should have received a copy of the Affero GNU General Public License      #
# along with PanACOTA (COPYING file).                                           #
# If not, see <https://www.gnu.org/licenses/>.                                  #
# ###############################################################################

"""
corepers is a subcommand of PanACoTA

Generate a core genome (families containing 1 member in all genomes of the dataset)
or a persistent genome (families with a given % of genomes having exactly 1 member).
You can also allow:

- mixed families: exactly 1 member in the given percentage of genomes, but the other genomes
  can contain 0 or several members
- multi families: allow several members in any genome.


@author gem
June 2017
"""

import os

import sys


def main_from_parse(args):
    """
    Call main function from the arguments given by parser

    Parameters
    ----------
    args : argparse.Namespace
        result of argparse parsing of all arguments in command line
    """
    cmd = "PanACoTA " + ' '.join(args.argv)
    main(cmd, args.pangenome, args.tol, args.multi, args.mixed, args.outputdir,
         floor=args.floor, verbose=args.verbose, quiet=args.quiet)


def main(cmd, pangenome, tol, multi, mixed, outputdir, floor=False, verbose=0, quiet=False):
    """
    Read pangenome and deduce Persistent genome according to the user criteria

    Parameters
    ----------
    pangenome : str
        file containing pangenome
    tol : float
        min % of genomes present in a family to consider it as persistent (between 0 and 1)
    multi : bool
        True if multigenic families are allowed, False otherwise
    mixed : bool
        True if mixed families are allowed, False otherwise
    outputdir : str or None
        Specific directory for the generated persistent genome. If not given, pangenome directory is used.
    floor : bool
        Require at least floor(nb_genomes*tol) genomes if True, ceil(nb_genomes*tol) if False
    verbose : int
        verbosity:
        - defaut 0 : stdout contains INFO, stderr contains ERROR.
        - 1: stdout contains INFO, stderr contains WARNING and ERROR
        - 2: stdout contains (DEBUG), DETAIL and INFO, stderr contains WARNING and ERROR
        - >=15: Add DEBUG in stdout
    quiet : bool
        True if nothing must be sent to stdout/stderr, False otherwise
    """
    # import needed packages
    import logging
    from PanACoTA import utils
    from PanACoTA import utils_pangenome as utilsp
    import PanACoTA.corepers_module.persistent_functions as pers
    from PanACoTA import __version__ as version

    # get pangenome name info
    _, base_pan = os.path.split(pangenome)
    # Define output filename
    output_name = "PersGenome_" + base_pan + "_"
    if floor:
        output_name += "F"
    output_name += str(tol)
    if multi:
        output_name += "-multi.lst"
    elif mixed:
        output_name += "-mixed.lst"
    else:
        output_name += ".lst"
    # Define output directory and filename path
    if not os.path.isdir(outputdir):
        os.makedirs(outputdir)
    outputfile = os.path.join(outputdir, output_name)
    logfile_base = os.path.join(outputdir, "PanACoTA-corepers")
    # level is the minimum level that will be considered.
    # for verbose = 0 or 1, ignore details and debug, start from info
    if verbose <= 1:
        level = logging.INFO
    # for verbose = 2, ignore only debug
    if verbose >= 2 and verbose < 15:
        level = 15 # int corresponding to detail level
    # for verbose >= 15, write everything
    if verbose >= 15:
        level = logging.DEBUG
    utils.init_logger(logfile_base, level, 'corepers', verbose=verbose, quiet=quiet)
    logger = logging.getLogger("corepers")
    logger.info(f'PanACoTA version {version}')
    logger.info("Command used\n \t > " + cmd)

    logger.info(get_info(tol, multi, mixed, floor))

    # Read pangenome
    fams_by_strain, families, all_strains = utilsp.read_pangenome(pangenome, logger)
    # Generate persistent genome
    fams = pers.get_pers(fams_by_strain, families, len(all_strains), tol, multi, mixed, floor)
    # Write persistent genome to file
    pers.write_persistent(fams, outputfile)
    logger.info("Persistent genome step done.")
    return outputfile


def get_info(tol, multi, mixed, floor):
    """
    Get a string corresponding to the information that will be given to logger.

    Parameters
    ----------
    tol : float
        min % of genomes present in a family to consider it as persistent (between 0 and 1)
    multi : bool
        True if multigenic families are allowed, False otherwise
    mixed : bool
        True if mixed families are allowed, False otherwise
    floor : bool
        Require at least floor(nb_genomes*tol) genomes if True, ceil(nb_genomes*tol) if False

    Returns
    -------
    str
        Information to give to logger
    """
    if tol == 1:
        return "Will generate a CoreGenome."
    else:
        if floor:
            floorstr = "floor"
        else:
            floorstr = "ceil"
        toprint = (f"Will generate a Persistent genome with member(s) in at least {100*tol}"
                   f"% of all genomes in each family.\n")
        if multi:
            toprint += ("Multigenic families are allowed (several members in "
                        "any genome of a family).")
        elif mixed:
            toprint += ("Mixed families are allowed. To be considered as persistent, "
                        f"a family must have exactly 1 member in {tol*100}% of the genomes, "
                        f"but in the remaining {round((1-tol)*100,3)}% genomes, there can be 0, 1 or "
                        "several members.")
        else:
            toprint += ("To be considered as persistent, a family must contain exactly 1 member "
                        f"in at least {tol*100}% of all genomes. The other genomes are absent from the "
                        "family.")
        return toprint


def build_parser(parser):
    """
    Method to create a parser for command-line options

    Parameters
    ----------
    parser : argparse.ArgumentParser
        parser to configure in order to extract command-line arguments
    """
    import argparse
    from PanACoTA import utils_argparse

    # Create command-line parser for all options and arguments to give
    required = parser.add_argument_group('Required arguments')
    required.add_argument("-p", dest="pangenome", required=True,
                          help="PanGenome file (1 line per family, first column is fam number)")
    required.add_argument("-o", dest="outputdir", required=True,
                          help=("Specify the output directory for your core/persistent genome."),
                          default=".")
    optional = parser.add_argument_group('Optional arguments')
    optional.add_argument("-t", "--tol", dest="tol", default=1, type=utils_argparse.percentage,
                          help=("min %% of genomes having at least 1 member in a family to "
                                "consider the family as persistent (between 0 and 1, "
                                "default is 1 = 100%% of genomes = Core genome)."
                                "By default, the minimum number of genomes will be "
                                "ceil('tol'*N) (N being the total number of genomes). If "
                                "you want to use floor('tol'*N) instead, add the '-F' option."))
    optional.add_argument("-M", dest="multi", action='store_true',
                          help=("Add this option if you allow several members in any genome "
                                "of a family. By default, only 1 (or 0 if tol<1) member "
                                "per genome are allowed in all genomes. If you want to allow "
                                "exactly 1 member in 'tol'%% of the genomes, and 0, 1 "
                                "or several members in the '1-tol'%% left, use the option -X "
                                "instead of this one: -M and -X options are not compatible."))
    optional.add_argument("-X", dest="mixed", action='store_true',
                          help="Add this option if you want to allow families having several "
                               "members only in '1-tol'%% of the genomes. In the other genomes, "
                               "only 1 member exactly is allowed. This option is not compatible "
                               "with -M (which is allowing multigenic families: having several "
                               "members in any number of genomes).")
    optional.add_argument("-F", dest="floor", action="store_true",
                          help="When you specify the '-tol' option, with a number lower "
                               "than 1, you can add this option to use floor('tol'*N) "
                               "as a minimum number of genomes instead of ceil('tol'*N) "
                               "which is the default behavior.")

    helper = parser.add_argument_group('Others')
    helper.add_argument("-v", "--verbose", dest="verbose", action="count", default=0,
                        help="Increase verbosity in stdout/stderr.")
    helper.add_argument("-q", "--quiet", dest="quiet", action="store_true", default=False,
                        help=("Do not display anything to stdout/stderr. log files will "
                              "still be created."))
    helper.add_argument("-h", "--help", dest="help", action="help",
                        help="show this help message and exit")


def check_args(parser, args):
    """
    Check that arguments given to parser are as expected.

    Parameters
    ----------
    parser : argparse.ArgumentParser
        The parser used to parse command-line
    args : argparse.Namespace
        Parsed arguments

    Returns
    -------
    argparse.Namespace or None
        The arguments parsed, updated according to some rules. Exit program
        with error message if error occurs with arguments given.
    """
    if args.multi and args.mixed:
        parser.error("-M and -X options cannot be activated together. Choose if you want to:\n"
                     "- allow several members in any number of genomes of a family (-M)\n"
                     "- allow several members in only '1-tol'% of the genomes of a family "
                     "(other 'tol'% genomes must have exactly 1 member) (-X)")
    if args.mixed and args.tol == 1:
        parser.error("You are asking for mixed families, while asking for 100% of the genomes of "
                     "a family to have exactly one member, which is not compatible. Do you want "
                     "to \n- lower the percentage of genomes required to have exactly "
                     "1 member (-t tol)\n- not allow mixed families (remove -X option)")
    if args.floor and args.tol == 1:
        parser.error("You are asking to use floor('tol'*N) as a minimum number of genomes "
                     "present in a family, but with 'tol'=1: the minimum number of genomes "
                     "will always be equal to N, using floor or the default ceil! Either "
                     "use a 'tol' lower than 1, or remove the '-F' option.")
    return args


def parse(parser, argu):
    """
    Parse arguments given to parser

    Parameters
    ----------
    parser : argparse.ArgumentParser
        the parser used
    argu : [str]
        command-line given by user, to parse using parser

    Returns
    -------
    argparse.Namespace
        Parsed arguments
    """
    args = parser.parse_args(argu)
    return check_args(parser, args)


if __name__ == '__main__':
    import argparse
    myparser = argparse.ArgumentParser(description="Compute core or persistent genome",
                                       add_help=False)
    build_parser(myparser)
    OPTIONS = parse(myparser, sys.argv[1:])
    main_from_parse(OPTIONS)



#!/usr/bin/env python3
# coding: utf-8

# ###############################################################################
# This file is part of PanACOTA.                                                #
#                                                                               #
# Authors: Amandine Perrin                                                      #
# Copyright © 2018-2020 Institut Pasteur (Paris).                               #
# See the COPYRIGHT file for details.                                           #
#                                                                               #
# PanACOTA is a software providing tools for large scale bacterial comparative  #
# genomics. From a set of complete and/or draft genomes, you can:               #
#    -  Do a quality control of your strains, to eliminate poor quality         #
# genomes, which would not give any information for the comparative study       #
#    -  Uniformly annotate all genomes                                          #
#    -  Do a Pan-genome                                                         #
#    -  Do a Core or Persistent genome                                          #
#    -  Align all Core/Persistent families                                      #
#    -  Infer a phylogenetic tree from the Core/Persistent families             #
#                                                                               #
# PanACOTA is free software: you can redistribute it and/or modify it under the #
# terms of the Affero GNU General Public License as published by the Free       #
# Software Foundation, either version 3 of the License, or (at your option)     #
# any later version.                                                            #
#                                                                               #
# PanACOTA is distributed in the hope that it will be useful, but WITHOUT ANY   #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS     #
# FOR A PARTICULAR PURPOSE. See the Affero GNU General Public License           #
# for more details.                                                             #
#                                                                               #
# You should have received a copy of the Affero GNU General Public License      #
# along with PanACOTA (COPYING file).                                           #
# If not, see <https://www.gnu.org/licenses/>.                                  #
# ###############################################################################

"""
Functions to:

- convert prokka tbl file to our tab file
- convert prokka ffn and faa headers to our format
- Create the database, with the following folders in the given "res_path":

    * Proteins: multifasta with all CDS in aa
    * Replicons: multifasta of genome
    * Genes: multifasta of all genes in nuc
    * gff3: gff files without sequence
    * LSTINFO: information on annotation. Columns are: "start end strand type locus\
    gene_name | product | EC_number | inference2" with the same types as prokka file,\
    and strain is C (complement) or D (direct). Locus is:\
    `<genome_name>.<contig_num><i or b>_<protein_num>`

@author gem
April 2019
"""

import os
import glob
import sys
import shutil
import logging
import PanACoTA.utils as utils
from PanACoTA.annotate_module import general_format_functions as general

logger = logging.getLogger("annotate.prokka_format")


def format_one_genome(gpath, name, prok_path, lst_dir, prot_dir, gene_dir,
                      rep_dir, gff_dir):
    """
     Format the given genome, and create its corresponding files in the following folders:

    - Proteins
    - Genes
    - Replicons
    - LSTINFO
    - gff

    Parameters
    ----------
    gpath : str
        path to the genome sequence which was given to prokka for annotation
    name : str
        gembase name of the genome
    prok_path : str
        directory where prokka folders are saved
    lst_dir : srt
        path to LSTINFO folder
    prot_dir : str
        path to Proteins folder
    gene_dir : str
        path to Genes folder
    rep_dir : str
        path to Replicons folder
    gff_dir : str
        path to gff3 folder

    Returns
    -------
    bool :
        True if genome was correctly formatted, False otherwise
    """
    prokka_dir = os.path.join(prok_path, os.path.basename(gpath) + "-prokkaRes")
    # Get needed Prokka result files
    fna_file = glob.glob(os.path.join(prokka_dir, "*.fna"))[0]
    prokka_tbl_file = glob.glob(os.path.join(prokka_dir, "*.tbl"))[0]
    prokka_gff_file = glob.glob(os.path.join(prokka_dir, "*.gff"))[0]
    prokka_ffn_file = glob.glob(os.path.join(prokka_dir, "*.ffn"))[0]
    prokka_faa_file = glob.glob(os.path.join(prokka_dir, "*.faa"))[0]

    #  Define names for generated gembase files
    res_lst_file = os.path.join(lst_dir, name + ".lst")
    res_gff_file = os.path.join(gff_dir, name + ".gff")
    res_gene_file = os.path.join(gene_dir, name + ".gen")
    res_prt_file = os.path.join(prot_dir, name + ".prt")
    res_rep_file = os.path.join(rep_dir, name + ".fna")


    # Generate replicon file (same as input sequence but with gembase formatted headers). From
    # this file, get contig names, to be used to generate gff file
    contigs, sizes = utils.get_genome_contigs_and_rename(name, fna_file, res_rep_file, logger)
    if not contigs:
        try:
            os.remove(res_rep_file)
            os.remove(res_lst_file)
            os.remove(res_gff_file)
            os.remove(res_gene_file)
            os.remove(res_prt_file)
        except OSError:
            pass
        logger.error("Problems while generating Replicon file for {}".format(name))
        return False

    # Convert prokka tbl file to gembase .lst file format
    ok_tbl = tbl2lst(prokka_tbl_file, res_lst_file, contigs, name, fna_file)
    if not ok_tbl:
        try:
            os.remove(res_rep_file)
            os.remove(res_lst_file)
            os.remove(res_gff_file)
            os.remove(res_gene_file)
            os.remove(res_prt_file)
        except OSError:
            pass
        logger.error("Problems while generating LSTINFO file for {}".format(name))
        return False
    # Create gff3 file for annotations
    ok_gff = generate_gff(fna_file, prokka_gff_file, res_gff_file, res_lst_file, sizes, contigs)
    if not ok_gff:
        try:
            os.remove(res_rep_file)
            os.remove(res_lst_file)
            os.remove(res_gff_file)
            os.remove(res_gene_file)
            os.remove(res_prt_file)
        except OSError:
            pass
        logger.error("Problems while generating .gff file for {}".format(name))
        return False
    # create Genes file (and check no problem occurred with return code)
    ok_gene = create_gen(prokka_ffn_file, res_lst_file, res_gene_file)
    # If gene file not created because a problem occurred, return False:
    # format did not run for this genome
    if not ok_gene:
        try:
            os.remove(res_rep_file)
            os.remove(res_lst_file)
            os.remove(res_gff_file)
            os.remove(res_gene_file)
            os.remove(res_prt_file)
        except OSError:
            pass
        logger.error("Problems while generating .gen file for {}".format(name))
        return False


    # If gene file was created, create Proteins file
    ok_prt = create_prt(prokka_faa_file, res_lst_file, res_prt_file)
    # If protein file not created, return False: format did not run for this genome
    if not ok_prt:
        try:
            os.remove(res_lst_file)
            os.remove(res_gff_file)
            os.remove(res_gene_file)
            os.remove(res_prt_file)
            os.remove(res_rep_file)
            # Remove twice to be able to check that when there is a problem while removing files,
            # it generates the expected error
            os.remove(res_rep_file)
        except OSError:
            pass
        logger.error("Problems while generating .prt file for {}".format(name))
        return False
    return True


def tbl2lst(tblfile, lstfile, contigs, genome, gpath):
    """
    Read prokka tbl file, and convert it to the lst file.

    * prokka tbl file format::

        >Feature contig_name
        start end type
                [EC_number text]
                [gene text]
                inference ab initio prediction:Prodigal:2.6
                [inference text]
                locus_tag test
                product text

    where `type` can be CDS, tRNA, rRNA, etc ...
    lines between [] are not always present

    * lst file format::

        start end strand type locus gene_name | product | EC_number | inference 2 | db_xref

    with the same types as prokka file, and strain is C (complement) or D (direct)
    locus is: `<genome_name>.<contig_num><i or b>_<protein_num>`

    Parameters
    ----------
    tblfile : str
        name of prokka output tbl file to read
    lstfile : str
        name of lst file to generate
    contigs : dict
        {original_contig_name: gembase_contig_name}
    genome : str
        genome name (gembase format)
    gpath : str
        path to the genome given to prodigal. Only used for error message
    changed_name : bool
        True if contig names have been changed (cutn != 0) -> contig names end by '_num',
        False otherwise.

    Returns
    -------
    bool :
        True if genome name used in lstfile and prokka tblfile are the same, False otherwise
    """
    # Protein localisation in contig (b = border ; i = inside)
    cont_loc = "b"
    prev_cont_loc = "b"
    # Current contig number. Used to compare with new one, to know if protein is
    # inside or at the border of a contig
    prev_cont_num = -1
    prev_cont_name = ""
    tbl_cont_name = ""
    tbl_cont_num = 0  # Contig number in tbl file (<orig_name>_<cont_num>)
    lst_cont_num = 0  # Exp cont num: new feature means exp_cont_num += 1
    # prev_cont_num = -1  # Previous contig number seen
    # Information on current feature. At the beginning, everything empty, no information
    gene_name = "NA"
    locus_num = "NA"
    product = "NA"
    ecnum = "NA"
    inf2 = "NA"
    db_xref = "NA"
    start = "-1"
    end = "-1"
    strand = "D"
    # Feature type (CDS, tRNA...)
    feature_type = ""

    # Check that tblfile is not empty
    if os.stat(tblfile).st_size == 0:
        logger.error(f"{tblfile} is empty.")
        return False

    with open(tblfile, "r") as tblf, open(lstfile, "w") as lstf:
        for line in tblf:
            elems = line.strip().split("\t")
            if line.startswith(">Feature"):
                tbl_cont_name = line.split()[-1]
            else:
                if not tbl_cont_name:
                    logger.error(f"Wrong format for {tblfile}.")
                    return False
                # Get line type, and retrieve info according to it
                # If it is not the line with start, end, type, there are only 2 elements
                #  in the line:
                #  - information_type information_value
                if len(elems) == 2:
                    if "locus_tag" in elems[0]:
                        locus_num = elems[-1].split("_")[-1]
                    if "gene" in elems[0]:
                        gene_name = elems[1]
                    if "product" in elems[0]:
                        product = elems[1]
                    if "EC_number" in elems[0]:
                        ecnum = elems[1]
                    if "inference" in elems[0] and "ab initio prediction:Prodigal" not in line:
                        inf2 = elems[1]
                    if "db_xref" in elems[0]:
                        db_xref = elems[1]
                # Information on start, end, type
                else:
                    # new gene
                    # if new gene is not on the same contig as previously,
                    # get new contig and loc = 'b'
                    if tbl_cont_name != prev_cont_name:
                        # Check that this contig name is in the list, and get its gembase contig
                        # number
                        if tbl_cont_name in contigs:
                            contig_num = contigs[tbl_cont_name].split(".")[-1]
                        # if not in the list, problem, return false
                        else:
                            logger.error(f"'{tbl_cont_name}' found in {tblfile} does not exist in "
                                         f"{gpath}.")
                            return False
                        # Previous loc was 'i' (because we were in the same contig as the
                        # previous one). But now, we know the it was the last gene of
                        # its contig: we change loc to 'b'
                        prev_cont_loc = "b"
                        cont_loc = "b"
                    # Same contig as previously: this gene is inside the contig (cont_loc = "i").
                    # If, in fact, it was the last gene of this contig, it will be changed
                    # when discovering next gene.
                    else:
                        cont_loc = 'i'

                    # If not first gene of the contig, write the previous gene to .lst file
                    # The first gene will be written while reading the 2nd gene
                    if start != "-1" and end != "-1" and not crispr:
                        lstline = general.write_gene(feature_type, locus_num,
                                                     gene_name, product,
                                                     prev_cont_loc, genome,
                                                     prev_cont_num, ecnum, inf2,
                                                     db_xref, strand, start, end, lstf)

                    # Get new values for the next gene: start, end, strand and feature type
                    start, end, feature_type = elems
                    crispr = "CRISPR" in feature_type or "repeat_region" in feature_type
                    if crispr:
                        continue

                    # Get strain of gene
                    if int(end) < int(start):
                        start, end = end, start
                        strand = "C"
                    else:
                        strand = "D"
                    # Initialize variables for next feature
                    # (except start, end, strand and feature type that we just calculated).
                    # prev_cont_num = exp_cont_num
                    prev_cont_loc = cont_loc
                    prev_cont_num = contig_num
                    prev_cont_name = tbl_cont_name
                    locus_num = "NA"
                    gene_name = "NA"
                    product = "NA"
                    ecnum = "NA"
                    inf2 = "NA"
                    db_xref = "NA"

        # Write last feature
        if start != -1 and end != -1:
            prev_cont_loc = "b"
            general.write_gene(feature_type, locus_num, gene_name, product,
                               prev_cont_loc, genome, prev_cont_num,
                               ecnum, inf2, db_xref, strand, start, end, lstf)
    return True


def generate_gff(gpath, prokka_gff_file, res_gff_file, res_lst_file, sizes, contigs):
    """
    From the lstinfo file and contig names (retrieved from generation of Replicons files),
    generate a gff file.

    Format:

    ##gff-version 3
    ##sequence-region contig1 start end
    ##sequence-region contig2 start end
    ...
    seqid(=contig) source   type start   end score  strand phase attributes

    All fields tab separated.
    Empty fields contain '.'

    For example:
    ESCO.1017.00200.00001    Prodigal:2.6    CDS start  end .   +   .   ID=ESCO.1017.00200.b0001_00001;locus_tag=ESCO.1017.00200.b0001_00001;product=hypothetical protein


    genome1_1   Prodigal_v2.6.3 CDS 213 1880    260.0   +   0   ID=1_1;partial=00;start_type=ATG;rbs_motif=None;rbs_spacer=None;gc_cont=0.534;conf=99.99;score=259.99;cscore=268.89;sscore=-8.89;rscore=-8.50;uscore=-4.34;tscore=3.95;


    Parameters
    ----------
    gpath : str
        path to the genome sequence given to prokka. Only used for error message
    res-gff_file : str
        path to the gff file that must be created in result database
    res-lst_file : str
        path to the lst file that was created in result database in the previous step
    sizes : list
        dict of contig names with their size. {"gembase1": "size", "gembase2":"size2" ...]
    contigs : list
        dict of contig original and gembase names. {"contig1": "gembase1"...}

    Returns
    -------
    bool :
        True if conversion worked well, False otherwise
    """
    # open gff generated by prokka to read it
    # open file to write new gff file (in gff3 folder)
    # open lst file (from LSTINFO folder) to read all annotation information saved
    # from prokka results
    with open(prokka_gff_file, "r") as prokf, open(res_lst_file, "r") as lstf, \
            open(res_gff_file, "w") as gfff:
        # Write headers of gff3 file
        gfff.write("##gff-version  3\n")
        # Write all sequences with their size. Order by name in gembase format
        for old, new in sorted(contigs.items(), key=lambda items:items[1]):
            end = sizes[new]
            # Write the list of cobisntigs, with their size
            gfff.write(f"##sequence-region\t{new}\t{1}\t{end}\n")

        # Now, convert each line of prokka gff to gembase formatted gff line
        for linegff in prokf:
            # Ignore gff lines starting by #. For the new gff, these are already all written at the
            # beginning of the file.
            if linegff.startswith("#"):
                continue
            # We do not write sequences
            if linegff.startswith('>'):
                break
            # Try to extract information on new line
            try:
                # Get all information from prokka gff line. Strip each of them as trailing whitespace
                # can be hidden (leading to information from gff considered as different from
                # information from lst)
                fields_g = linegff.strip().split("\t")
                fields_g = [info.strip() for info in fields_g]
                # Ignore lines with sequence
                # if len(fields_g) != 9:
                #     continue
                (contig_name, source, type_g, start_g, end_g,
                 score, strand_g, phase, attributes) = fields_g
                # Ignore CRISPR
                if "CRISPR" in type_g or "repeat_region" in type_g:
                    continue
                # Get information given to this same sequence from the lst file
                # (next lst line corresponds to next gff line without #), as, for each format,
                # there is 1 line per gene)
                linelst = lstf.readline()
                fields_l = linelst.split("\t")
                fields_l = [info.strip() for info in fields_l]
                start_l, end_l, strand_l, type_l, locus_l, l_gene, l_info = fields_l
                # Get gff and ffn filenames to give information to user if error message
                gff = os.path.basename(prokka_gff_file)
                tbl = gff.replace(".gff", ".tbl")
                # Path where gff and ffn generated by prodigal are
                tmp = gpath + "-prokkaRes"
                # Get gene name given by prodigal to current gene
                gname = attributes.split("ID=")[1].split(";")[0]
                # Get locus_tag given by prokka to current feature (should be the same as ID)
                loc_name = attributes.split("locus_tag=")[1].split(";")[0]
                if loc_name != gname:
                    logger.error(f"Problem in {gff}: ID={gname} whereas locus_tag={loc_name}.")
                    return False
                # Compare information from lst and information from prodigal gff (start,
                # end and type of feature). They should correspond
                for (elemg, eleml, label) in zip([start_g, end_g, type_g],
                                                 [start_l, end_l, type_l],
                                                 ["start", "end", "type"]):
                    # If 1 element is different (start or end position, or type), print error
                    # message and return False: this genome could not be converted to gff
                    if elemg != eleml:
                        logger.error(f"Files {tbl} and {gff} (in prokka tmp_files: {tmp}) "
                                     f"do not have the same {label} value for gene {gname} ({elemg} "
                                     f"in gff, {eleml} in tbl)")
                        return False

                # Replace prokka ID and locus_tag with the new gene name (in gembase format),
                # found in the corresponding lst line
                at_split = attributes.split(";")
                newID = [atr if "ID" not in atr else f'ID={locus_l}' for atr in at_split]
                new = [atr if "locus_tag" not in atr else f'locus_tag={locus_l}' for atr in newID]

                # Write new line of gff file
                cname = contigs[contig_name]
                info = "\t".join([cname, source, type_g, start_g, end_g, score, strand_g,
                                  phase, ";".join(new)])
                gfff.write(info + "\n")
            except:
                logger.error(f"Wrong format for {prokka_gff_file}.")
                return False
        return True


def create_gen(ffnseq, lstfile, genseq):
    """
    Generate .gen file, from sequences contained in .ffn, but changing the
    headers using the information in .lst

    Parameters
    ----------
    ffnseq : str
        .ffn file generated by prokka
    lstfile : str
        lstfile converted from prokka tbl file
    genseq : str
        output file, to write in Genes directory
    logger : logging.Logger
        logger object to put information

    Returns
    -------
    bool :
        True if conversion went well, False otherwise
    """
    problem = False
    write = True  # Write next sequence
    with open(ffnseq) as ffn, open(lstfile) as lst, open(genseq, "w") as gen:
        for line_ffn in ffn:
            # Ignore gene that we do not want to write (should be a crispr)
            # If line of sequence, write it as is, and go to next line
            if not line_ffn.startswith(">"):
                # We just read a seq line. If we can write (write is True), do it and go
                # to next line
                # Otherwise, just go to next line
                if write:
                    gen.write(line_ffn)
                continue
            # Try to get gene ID. If does not work, ignore this gene (it may be a
            # CRISPR, and we ignore them
            test_gen_id = line_ffn.split()[0].split("_")[-1]
            if not test_gen_id.isdigit():
                # Maybe a CRISPR? Or wrong gene name? -> ignore
                logger.log(utils.detail_lvl(),
                           f"Unknown header format for {line_ffn.strip()}. "
                           "This gene will be ignored in .gen output file.")
                write = False
                continue
            # If ffn contains a gene header, find its information in lst file
            else:
                write = True
                lstline = lst.readline().strip()
                gen_id = int(test_gen_id)
                # genID exists, ffn header is for a gene. Check that it corresponds to
                # information in lst file.
                id_lst = lstline.split("\t")[4].split("_")[-1]
                gen_id_lst = int(id_lst)
                # in lst, find the same gene ID as in ffn (some gene IDs in lst can be absent
                # from ffn, if prokka do not give their sequence).
                # As they are ordered by increasing number, go to next lstline until
                # corresponding gene ID is found. However, if ffn ID > lst ID: ID does not
                # exist in .lst -> problem.
                while gen_id > gen_id_lst:
                    lstline = lst.readline().strip()
                    if not lstline:
                        gen_id_lst = "-1"
                        break
                    id_lst = lstline.split("\t")[4].split("_")[-1]
                    gen_id_lst = int(id_lst)
                # If it found the same gene ID, write info in gene file
                if gen_id == gen_id_lst:
                    general.write_header(lstline.strip(), gen)
                # If gene ID of ffn not found, write error message and stop
                else:
                    logger.error(f"Missing info for gene {line_ffn.strip()} "
                                 f"(from {ffnseq}) in {lstfile}. If it is actually present "
                                 "in the lst file, check that genes are ordered by increasing number in both lst and ffn files.")
                    return False
    return True


def create_prt(faaseq, lstfile, prtseq):
    """
    Generate .prt file, from sequences in .faa, but changing the headers
    using information in .lst

    **Note:** works if proteins are in increasing order (of number after "_" in their name)
    in faa and tbl (hence lst) files.

    If a header is not in the right format, or a protein exists in prt file but not in lstfile,
    conversion is stopped, an error message is output, and prt file is removed.

    Parameters
    ----------
    faaseq : str
        faa file output of prokka
    lstfile : str
        lstinfo converted from prokka tab file
    prtseq : str
        output file where converted proteins must be saved

    Returns
    -------
    bool :
        True if conversion went well, False otherwise
    """
    problem = False
    with open(faaseq) as faa, open(lstfile) as lst, open(prtseq, "w") as prt:
        for line in faa:
            # all header lines must start with PROKKA_<geneID>
            if line.startswith(">"):
                try:
                    # get gene ID
                    gen_id = int(line.split()[0].split("_")[-1])
                except ValueError as err:
                    logger.error(f"Unknown header format {line.strip()} in {faaseq}. "
                                 f"Gene ID is not a number.")
                    return False
                gen_id_lst = 0
                # get line of lst corresponding to the gene ID
                lstline = ""
                while gen_id > gen_id_lst:
                    lstline = lst.readline().strip()
                    id_lst = lstline.split("\t")[4].split("_")[-1]
                    # don't cast to int if info for a crispr
                    if id_lst.isdigit():
                        gen_id_lst = int(id_lst)
                # check that gen_id is the same as the lst line
                if gen_id == gen_id_lst:
                    general.write_header(lstline, prt)
                else:
                    logger.error(f"Missing info for protein {line.strip()} (from {faaseq}) "
                                 f"in {lstfile}. If it is actually present "
                                 "in the lst file, check that proteins are ordered by increasing "
                                 "number in both lst and faa files.")
                    return False
            # not header: inside sequence, copy it to the .prt file
            else:
                prt.write(line)
    return True





#!/usr/bin/env python3
# coding: utf-8

# ###############################################################################
# This file is part of PanACOTA.                                                #
#                                                                               #
# Authors: Amandine Perrin                                                      #
# Copyright © 2018-2020 Institut Pasteur (Paris).                               #
# See the COPYRIGHT file for details.                                           #
#                                                                               #
# PanACOTA is a software providing tools for large scale bacterial comparative  #
# genomics. From a set of complete and/or draft genomes, you can:               #
#    -  Do a quality control of your strains, to eliminate poor quality         #
# genomes, which would not give any information for the comparative study       #
#    -  Uniformly annotate all genomes                                          #
#    -  Do a Pan-genome                                                         #
#    -  Do a Core or Persistent genome                                          #
#    -  Align all Core/Persistent families                                      #
#    -  Infer a phylogenetic tree from the Core/Persistent families             #
#                                                                               #
# PanACOTA is free software: you can redistribute it and/or modify it under the #
# terms of the Affero GNU General Public License as published by the Free       #
# Software Foundation, either version 3 of the License, or (at your option)     #
# any later version.                                                            #
#                                                                               #
# PanACOTA is distributed in the hope that it will be useful, but WITHOUT ANY   #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS     #
# FOR A PARTICULAR PURPOSE. See the Affero GNU General Public License           #
# for more details.                                                             #
#                                                                               #
# You should have received a copy of the Affero GNU General Public License      #
# along with PanACOTA (COPYING file).                                           #
# If not, see <https://www.gnu.org/licenses/>.                                  #
# ###############################################################################

"""
Functions used to deal with pangenome file

@author gem
April 2017
"""
import logging
import os
from PanACoTA import utils

logger = logging.getLogger("utils.pan")


def read_pangenome(pangenome, logger, families=None):
    """
    Read pangenome information

    Read pangenome according to what is available. First, check if python objects are available,
    then if not, search for the binary file, and if not, read the text file.

    Parameters
    ----------
    pangenome : str
        path to pangenome file
    families : dict or None
        {num: [members]} if families are given. If not (must read them from binary file if exists\
        or pangenome file otherwise), None.

    Returns
    -------
    (fams_by_strain, families, all_strains) : tuple
        with:

        - fams_by_strain: {fam_num: {strain: [members]}}
        - families: {fam_num: [all members]}
        - all_strains: list of all genome names
    """
    if families:
        fams_by_strain, all_strains = get_fams_info(families, logger)
        if not os.path.isfile(pangenome + ".bin"):
            logger.details("Saving all information to a binary file for later use")
            utils.save_bin([fams_by_strain, families, all_strains], pangenome + ".bin")
    elif os.path.isfile(pangenome + ".bin"):
        logger.info("Retrieving info from binary file")
        fams_by_strain, families, all_strains = utils.load_bin(pangenome + ".bin")
    else:
        fams_by_strain, families, all_strains = read_pan_file(pangenome, logger)
        logger.info("Saving all information to a binary file for later use")
        utils.save_bin([fams_by_strain, families, all_strains], pangenome + ".bin")
    return fams_by_strain, families, all_strains


def get_fams_info(families, logger):
    """
    From all families as list of members, get more information:

    - all strains found, sorted by species name
    - for each family, sort members by strain

    Parameters
    ----------
    families : dict
        {num: [members]}
    logger : logging.Logger
        logger object to write log information

    Returns
    -------
    (fams_by_strain, sorted_all_strains) : tuple
        with:

        - fams_by_strain: {fam_num: {strain: [members], strain: [members]}}
        - sorted_all_strains: list of all strains found, sorted by species
    """
    logger.info("Retrieving information from pan families")
    fams_by_strain = {}
    all_strains = set()
    for num, fam in families.items():
        fams_by_strain[num] = {}
        for gene in fam:
            read_gene(gene, num, fams_by_strain, all_strains)
    sort_all_strains = sorted(list(all_strains), key=utils.sort_genomes_by_name)
    return fams_by_strain, sort_all_strains


def read_pan_file(filein, logger):
    """
    Read PanGenome file in 'filein', and put it into Python objects

    Parameters
    ----------
    filein : str
        path to pangenome file
    logger :

    Returns
    -------
    (fams_by_strain, families, sort_all_strains) : tuple
        with:

        - fams_by_strain: {fam_num: {strain: [members]}}
        - families: {fam_num: [all members]}
        - sort_all_strains: list of all genome names, sorted by species name
    """
    logger.info("Reading and getting information from pangenome file")
    fams_by_strain = {}
    families = {}
    all_strains = set()
    nfam = 0
    with open(filein, 'r') as coref:
        for line in coref:
            genes = line.strip().split()
            fam_num = genes[0]
            fams_by_strain[fam_num] = {}
            genes_ok = genes[1:]
            for gene in genes_ok:
                read_gene(gene, fam_num, fams_by_strain, all_strains)
            families[fam_num] = genes_ok
            nfam += 1
    sort_all_strains = sorted(list(all_strains), key=utils.sort_genomes_by_name)
    return fams_by_strain, families, sort_all_strains


def read_gene(gene, num, fams_by_strain, all_strains):
    """
    Read information from a given gene name, and save it to appropriate dicts

    Parameters
    ----------
    gene : str
        gene name (species.date.strain.contig_number
    num : str
        num of family from which the given gene is
    fams_by_strain : dict
        {fam_num: {strain: [members]}}
    all_strains : set
        set of all strains

    """
    # if format is ESCO.1512.00001.i001_12313 genome name is ESCO.1512.00001
    if "." in gene and len(gene.split(".")) >= 3:
        strain = ".".join(gene.split("_")[0].split(".")[:3])
    # otherwise, genename is everything before the last "_"
    else:
        strain = "_".join(gene.split("_")[:-1])
    if strain in fams_by_strain[num]:
        fams_by_strain[num][strain].append(gene)
    else:
        fams_by_strain[num][strain] = [gene]
    if strain not in all_strains:
        all_strains.add(strain)



#!/usr/bin/env python3
# coding: utf-8

"""
Functional tests for the parser of 'all' subcommand
"""
import argparse
import pytest

from PanACoTA.subcommands import all_modules as allm


def test_parser_noarg(capsys):
    """
    Test that when the script is called without any argument, an error message appears,
    indicating the required arguments.
    """
    parser = argparse.ArgumentParser(description="Run all modules", add_help=False)
    allm.build_parser(parser)
    with pytest.raises(SystemExit):
        allm.parse(parser, "".split())
    _, err = capsys.readouterr()
    assert "usage: " in err
    assert "-o OUTDIR [--threads THREADS]" in err
    assert "[-t NCBI_SPECIES_TAXID]" in err
    assert "-o OUTDIR" in err
    assert "[-s NCBI_SPECIES]" in err
    assert "[-l LEVELS]" in err
    assert "[--cutn CUTN] [--l90 L90]" in err
    assert "[--tol TOL]" in err
    assert "[-Mu] [-X]" in err
    assert "[--soft {fasttree,fastme,quicktree,iqtree,iqtree2}]" in err
    assert "the following arguments are required: -o, -n" in err


def test_parser_noconffile():
    """
    Test running with arguments ok for prepare, no configfile given.
    Only args given in line change, all others are by default
    """
    parser = argparse.ArgumentParser(description="Run all modules", add_help=False)
    allm.build_parser(parser)
    options = allm.parse(parser, "-o out-all -n TEST -t 1234".split())
    assert options.outdir == "out-all"
    assert options.threads == 1
    assert options.ncbi_species_taxid == '1234'
    assert options.prodigal_only == False
    assert not options.norefseq
    assert options.l90 == 100
    assert options.clust_mode == 1
    assert options.min_id == 0.8
    assert options.tol == 1
    assert not options.boot


def test_parser_conffile():
    """
    Test running with arguments with config file.
    arguments given in cmd line are kept, others are those in configfile,
    and the ones not defined in configfile are by default
    """
    parser = argparse.ArgumentParser(description="Run all modules", add_help=False)
    allm.build_parser(parser)
    options = allm.parse(parser,
                         "-c test/data/all/default-conffigfile.ini -o out-all -n TEST".split())
    assert options.outdir == "out-all"
    assert options.threads == 10
    assert not options.ncbi_species_taxid
    assert options.prodigal_only == False
    assert options.norefseq
    assert options.l90 == 99
    assert options.clust_mode == 1
    assert options.min_id == 0.85
    assert options.tol == 1
    assert not options.multi
    assert not options.boot


def test_parser_conffile_and_cmd():
    """
    Test that when some arguments given in config file are also given in cmd,
    the value kept is the one in cmd.
    """
    import multiprocessing
    nb = multiprocessing.cpu_count()
    parser = argparse.ArgumentParser(description="Run all modules", add_help=False)
    allm.build_parser(parser)
    options = allm.parse(parser,
                         "-c test/data/all/default-conffigfile.ini -o out-all "
                         "-n TEST --threads 0 -i 0.99 -Mu".split())
    assert options.outdir == "out-all"
    assert options.threads == nb
    assert not options.ncbi_species_taxid
    assert options.prodigal_only == False
    assert options.norefseq
    assert options.l90 == 99
    assert options.clust_mode == 1
    assert options.min_id == 0.99
    assert options.tol == 1
    assert options.multi
    assert not options.boot


def test_parser_thread_quicktree(capsys):
    """
    Testing default that quicktree + threads != 1 gives an error
    """
    parser = argparse.ArgumentParser(description="Run all modules", add_help=False)
    allm.build_parser(parser)
    with pytest.raises(SystemExit):
        allm.parse(parser,
                         "-c test/data/all/default-conffigfile.ini -o out-all "
                         "-n TEST --threads 2 --soft quicktree".split())
    _, err = capsys.readouterr()
    assert ("You cannot run quicktree with multiple threads. "
            "Choose another software, or remove the --threads option") in err


def test_duplicate_value(capsys):
    """
    test that when a parameter is given twice in the configfile, it raises an error
    """
    parser = argparse.ArgumentParser(description="Run all modules", add_help=False)
    allm.build_parser(parser)
    with pytest.raises(SystemExit):
        allm.parse(parser,
                         "-c test/data/all/conffile-duplicate.ini -o out-all "
                         "-n TEST --threads 2 --soft quicktree".split())
    out, err = capsys.readouterr()
    assert ("option 'verbose' in section 'prepare' already exists") in out


def test_wrong_conffile_name(capsys):
    """
    test that when the given config file does not exist, it raises an error
    """
    parser = argparse.ArgumentParser(description="Run all modules", add_help=False)
    allm.build_parser(parser)
    with pytest.raises(SystemExit):
        allm.parse(parser, "-c conf.ini -o out-all -t 1234 -n TEST".split())
    out, err = capsys.readouterr()
    assert ("Error: config file conf.ini not found.") in out


def test_annot_defvalue(capsys):
    """
    Test when a value is given in 'default' section of config file, but redefined in
    another section, it keeps the redefined value
    """
    parser = argparse.ArgumentParser(description="Run all modules", add_help=False)
    allm.build_parser(parser)
    options = allm.parse(parser,
                         "-c test/data/all/default-conffigfile.ini -o out-all -n TEST".split())
    assert options.outdir == "out-all"
    assert options.threads == 10
    assert not options.ncbi_species_taxid
    assert options.prodigal_only == False
    assert options.min_dist == '5'
    assert options.norefseq
    assert options.l90 == 99
    assert options.clust_mode == 1
    assert options.min_id == 0.85
    assert options.tol == 1
    assert not options.multi
    assert not options.boot





#!/usr/bin/env python3

# ###############################################################################
# This file is part of PanACOTA.                                                #
#                                                                               #
# Authors: Amandine Perrin                                                      #
# Copyright © 2018-2020 Institut Pasteur (Paris).                               #
# See the COPYRIGHT file for details.                                           #
#                                                                               #
# PanACOTA is a software providing tools for large scale bacterial comparative  #
# genomics. From a set of complete and/or draft genomes, you can:               #
#    -  Do a quality control of your strains, to eliminate poor quality         #
# genomes, which would not give any information for the comparative study       #
#    -  Uniformly annotate all genomes                                          #
#    -  Do a Pan-genome                                                         #
#    -  Do a Core or Persistent genome                                          #
#    -  Align all Core/Persistent families                                      #
#    -  Infer a phylogenetic tree from the Core/Persistent families             #
#                                                                               #
# PanACOTA is free software: you can redistribute it and/or modify it under the #
# terms of the Affero GNU General Public License as published by the Free       #
# Software Foundation, either version 3 of the License, or (at your option)     #
# any later version.                                                            #
#                                                                               #
# PanACOTA is distributed in the hope that it will be useful, but WITHOUT ANY   #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS     #
# FOR A PARTICULAR PURPOSE. See the Affero GNU General Public License           #
# for more details.                                                             #
#                                                                               #
# You should have received a copy of the Affero GNU General Public License      #
# along with PanACOTA (COPYING file).                                           #
# If not, see <https://www.gnu.org/licenses/>.                                  #
# ###############################################################################

"""
Functions helping for downloading refseq genomes of a species,
gunzip them, adding complete genomes...

@author gem
August 2017
"""
import os
import logging
import shutil
import sys
import glob
import urllib.request
import ncbi_genome_download as ngd

from PanACoTA import utils


logger = logging.getLogger("prepare.dds")


def download_from_ncbi(species_linked, section, ncbi_species_name, 
    ncbi_species_taxid, ncbi_taxid, levels, outdir, threads):
    """
    Download ncbi genomes of given species

    Parameters
    ----------
    species_linked : str
        given NCBI species with '_' instead of spaces, or NCBI taxID if species
        name not given
    section : str
        genbank or only refseq (default = refseq)
    ncbi_species_name : str or None
        name of species to download: user given NCBI species. None if
        no species name given
    ncbi_species_taxid : int
        species taxid given in NCBI (-T option)
    ncbi_taxid : int
        taxid given in NCBI (-t option)
    outdir : str
        Directory where downloaded sequences must be saved
    threads : int
        Number f threads to use to download genome sequences

    Returns
    -------
    str :
        Output filename of downloaded summary

    """
    # Name of summary file, with metadata for each strain:
    sumfile = os.path.join(outdir, f"assembly_summary-{species_linked}.txt")
    abs_sumfile = os.path.abspath(sumfile)

    # arguments needed to download all genomes of the given species
    abs_outdir = os.path.abspath(outdir)
    keyargs = {"section": section, "file_formats": "fasta", 
               "output": abs_outdir,
               "parallel": threads, "groups": "bacteria",
               "metadata_table":abs_sumfile}
    message = "Downloading all genomes for "
    # If NCBI species given, add it to arguments to download genomes, and write it to info message
    if ncbi_species_name:
        keyargs["genera"] = ncbi_species_name
        message += f"NCBI species = {ncbi_species_name}"
    # If NCBI species given, add it to arguments to download genomes, and write it to info message
    if ncbi_species_taxid:
        keyargs["species_taxids"] = ncbi_species_taxid
        if ncbi_species_name:
            message += f" (NCBI_species_taxid = {ncbi_species_taxid})."
        else:
            message += f" NCBI_species_taxid = {ncbi_species_taxid}"
    if ncbi_taxid:
        keyargs["taxids"] = ncbi_taxid
        if ncbi_species_name or ncbi_species_taxid:
            message += f" (and NCBI_taxid = {ncbi_taxid})."
        else:
            message += f" NCBI_taxid = {ncbi_taxid}"

    # If assembly level(s) given, add it to arguments, and write to info message
    if levels:
        keyargs["assembly_levels"] = levels
        message += f" (Only those assembly levels: {levels}). "
    logger.info(f"Metadata for all genomes will be saved in {sumfile}")
    logger.info(message)

    # Download genomes
    max_retries = 15 # If connection to NCBI fails, how many retry downloads must be done
    error_message = ("Could not download genomes. Check that you gave valid NCBI taxid and/or "
                     "NCBI species name. If you gave both, check that given taxID and name really "
                     "correspond to the same species.")
    # widgets = [progressbar.BouncingBar(marker=progressbar.RotatingMarker(markers="◐◓◑◒")),
    #            "  -  ", progressbar.Timer()]
    # bar = progressbar.ProgressBar(widgets=widgets, max_value=20, term_width=50)
    try:
        # Download genomes
        # ret = None
        # while True:
        #     if ret:
        #         break
        #     bar.update()
        ret = ngd.download(**keyargs)

    except: # pragma: no cover
        # Error message if crash during execution of ncbi_genome_download
        logger.error(error_message)
        # bar.finish()
        sys.exit(1)
    attempts = 0
    while ret == 75 and attempts < max_retries: # pragma: no cover
        # bar.update()
        attempts += 1
        logging.error(('Downloading from NCBI failed due to a connection error, '
                       'retrying. Already retried so far: %s'), attempts)
        ret = ngd.download(**keyargs)
    # bar.finish()
    # Message if NGD did not manage to download the genomes (wrong species name/taxid)
    if ret != 0:
        # Error message
        logger.error(error_message)
        sys.exit(1)
    nb_gen, db_dir = to_database(outdir, section)
    return db_dir, nb_gen


def to_database(outdir, section):
    """
    Move .fna.gz files to 'database_init' folder, and uncompress them.

    Parameters
    ----------
    outdir : str
        directory where all results are (for now, refseq/genbank folders, assembly summary and log
    section : str
        refseq (default) or genbank

    Returns
    -------
        nb_gen : number of genomes downloaded
        db_dir : directory where are all fna files downloaded from refseq/genbank
    """
    # Copy .gz files in a new folder, and Unzip them in this new folder
    logger.info("Uncompressing genome files.")
    # Folder where are .gz files
    download_dir = os.path.join(outdir, section, "bacteria")
    # If no folder output/refseq/bacteria: error, no genome found
    # (or output/genbank/bacteria)
    if not os.path.exists(download_dir):
        logger.error(f"The folder containing genomes downloaded from NCBI {section} "
                     f"({download_dir}) does not exist. Check that you really downloaded "
                     "sequences (fna.gz) and that they are in this folder.")
        sys.exit(1)
    # If folder output/<refseq or genbank>/bacteria empty: error, no genome found
    list_downloads = os.listdir(download_dir)
    if list_downloads == []:
        logger.error(f"The folder supposed to contain genomes downloaded from NCBI {section} "
                     f"({download_dir}) exists but is empty. Check that you really downloaded "
                     "sequences (fna.gz).")
        sys.exit(1)
    # Create directory to put uncompressed genomes
    db_dir = os.path.join(outdir, "Database_init")
    os.makedirs(db_dir, exist_ok=True)
    nb_gen = 0
    # For each subfolder of download dir, move the .gz file it contains (if possible)
    # to the new database folder
    for g_folder in os.listdir(download_dir):
        fasta = glob.glob(os.path.join(download_dir, g_folder, "*.fna.gz"))
        # No .gz file in folder
        if len(fasta) == 0:
            logger.warning("Problem with genome in {}: no compressed fasta file downloaded. "
                           "This genome will be ignored.".format(g_folder))
            continue
        # Several gz files in folder
        elif len(fasta) > 1:
            logger.warning("Problem with genome in {}: several compressed fasta files found. "
                           "This genome will be ignored.".format(g_folder))
            continue
        # Copy gz file to new folder
        fasta_file = os.path.basename(fasta[0])
        fasta_out = os.path.join(db_dir, fasta_file)
        shutil.copy(fasta[0], fasta_out)
        # Uncompress file copied
        cmd = f"gunzip {fasta_out} -f"
        error = f"Error while trying to uncompress {fasta_out}. This genome will be ignored."
        call = utils.run_cmd(cmd, error)
        # Problem with uncompressing: genome ignored (remove gz file from new folder)
        if call.returncode != 0:
            os.remove(fasta_out)
            continue
        nb_gen += 1
    return nb_gen, db_dir



#!/usr/bin/env python3
# coding: utf-8

"""
Unit tests for the utils_pangenome submodule of PanACoTA
"""
import logging
import os
import shutil
import pytest

from PanACoTA import utils_pangenome as upan
from PanACoTA import utils


# Define functions and variables shared by several tests
def my_logger():
    """
    logger given to function called by a subprocess
    """

    def make_logger(name="test_post_mmseq"):
        """
        Create logger according to name given
        """
        logfile_base = "log_" + name
        level = logging.DEBUG
        utils.init_logger(logfile_base, level, name, verbose=0, quiet=False)
        return logfile_base

    return make_logger

# Define common variables
PAN_PATH = os.path.join("test", "data", "pangenome")
PAN_EXP = os.path.join(PAN_PATH, "exp_files")
PAN_TEST = os.path.join(PAN_PATH, "test_files")
GENEPATH = os.path.join(PAN_PATH, "generated_by_unit-tests")
PAN_FILE = os.path.join(PAN_EXP, "exp_pangenome-4genomes.lst")
LOGFILE_BASE = "test_getseqs-logs"
LOG_LEVEL = logging.DEBUG
LOGFILES = [LOGFILE_BASE + ext for ext in [".log", ".log.debug", ".log.details", ".log.err"]]

@pytest.fixture(autouse=True)
def setup_teardown_module():
    """
    Remove log files at the end of this test module

    Before each test:
    - init logger
    - create directory to put generated files

    After:
    - remove all log files
    - remove directory with generated results
    """
    utils.init_logger(LOGFILE_BASE, LOG_LEVEL, 'test_utilspan', verbose=1)
    os.mkdir(GENEPATH)
    print("setup")

    yield
    for f in LOGFILES:
        if os.path.exists(f):
            os.remove(f)
    shutil.rmtree(GENEPATH)
    print("teardown")



FAMS_BY_STRAIN = \
    {'1': {'GEN2.1017.00001': ['GEN2.1017.00001.i0002_00004'],
           'GEN4.1111.00001': ['GEN4.1111.00001.i0001_00002'],
           'GENO.1017.00001': ['GENO.1017.00001.b0002_00003'],
           'GENO.1216.00002': ["GENO.1216.00002.i0001_00003"]
           },
     '2': {'GEN2.1017.00001': ['GEN2.1017.00001.b0003_00010']},
     '3': {'GEN2.1017.00001': ['GEN2.1017.00001.b0004_00013']},
     '4': {'GEN2.1017.00001': ['GEN2.1017.00001.i0002_00005'],
           'GEN4.1111.00001': ["GEN4.1111.00001.b0001_00001"],
           'GENO.1017.00001': ["GENO.1017.00001.b0001_00002"],
           'GENO.1216.00002': ["GENO.1216.00002.b0001_00001", "GENO.1216.00002.i0001_00002"]},
     '5': {'GEN4.1111.00001': ['GEN4.1111.00001.i0001_00003']},
     '6': {'GEN2.1017.00001': ['GEN2.1017.00001.b0004_00011'],
           'GEN4.1111.00001': ['GEN4.1111.00001.b0001_00009'],
           'GENO.1017.00001': ['GENO.1017.00001.b0002_00011'],
           'GENO.1216.00002': ['GENO.1216.00002.b0002_00010']
           },
     '7': {'GEN2.1017.00001': ['GEN2.1017.00001.b0002_00006'],
           'GENO.1017.00001': ["GENO.1017.00001.b0001_00001"]},
     '8': {'GEN2.1017.00001': ['GEN2.1017.00001.b0003_00007'],
           'GEN4.1111.00001': ['GEN4.1111.00001.i0001_00006'],
           'GENO.1017.00001': ['GENO.1017.00001.i0002_00006', 'GENO.1017.00001.i0002_00007'],
           'GENO.1216.00002': ['GENO.1216.00002.i0001_00007']
           },
     '9': {'GEN2.1017.00001': ['GEN2.1017.00001.i0004_00012'],
           'GENO.1017.00001': ['GENO.1017.00001.i0002_00008']},
     '10': {'GEN2.1017.00001': ['GEN2.1017.00001.i0003_00008'],
            'GEN4.1111.00001': ['GEN4.1111.00001.i0001_00007'],
            'GENO.1017.00001': ['GENO.1017.00001.i0002_00009'],
            'GENO.1216.00002': ['GENO.1216.00002.b0001_00008']
            },
     '11': {'GEN2.1017.00001': ['GEN2.1017.00001.i0003_00009'],
            'GEN4.1111.00001': ['GEN4.1111.00001.i0001_00008'],
            'GENO.1017.00001': ['GENO.1017.00001.i0002_00010'],
            'GENO.1216.00002': ['GENO.1216.00002.b0002_00009']
            },
     '12': {'GEN2.1017.00001': ['GEN2.1017.00001.b0002_00003'],
            'GENO.1216.00002': ['GENO.1216.00002.i0001_00004']
            },
     '13': {'GEN2.1017.00001': ['GEN2.1017.00001.b0001_00002'],
            'GEN4.1111.00001': ['GEN4.1111.00001.i0001_00004'],
            'GENO.1017.00001': ['GENO.1017.00001.i0002_00004'],
            'GENO.1216.00002': ['GENO.1216.00002.i0001_00005']},
     '14': {'GEN2.1017.00001': ['GEN2.1017.00001.b0001_00001'],
            'GEN4.1111.00001': ['GEN4.1111.00001.i0001_00005'],
            'GENO.1017.00001': ['GENO.1017.00001.i0002_00005'],
            'GENO.1216.00002': ['GENO.1216.00002.i0001_00006']},
     '15': {'GENO.1216.00002': ['GENO.1216.00002.b0003_00011']},
     '16': {'GENO.1216.00002': ['GENO.1216.00002.b0003_00012']}
     }

FAMILIES = {'1': ['GEN2.1017.00001.i0002_00004', 'GEN4.1111.00001.i0001_00002',
                  'GENO.1017.00001.b0002_00003', 'GENO.1216.00002.i0001_00003'],
            '2': ['GEN2.1017.00001.b0003_00010'],
            '3': ['GEN2.1017.00001.b0004_00013'],
            '4': ['GEN2.1017.00001.i0002_00005', 'GEN4.1111.00001.b0001_00001',
                  'GENO.1017.00001.b0001_00002', 'GENO.1216.00002.b0001_00001',
                  'GENO.1216.00002.i0001_00002'],
            '5': ['GEN4.1111.00001.i0001_00003'],
            '6': ['GEN2.1017.00001.b0004_00011', 'GEN4.1111.00001.b0001_00009',
                  'GENO.1017.00001.b0002_00011', 'GENO.1216.00002.b0002_00010'],
            '7': ['GEN2.1017.00001.b0002_00006', 'GENO.1017.00001.b0001_00001'],
            '8': ['GEN2.1017.00001.b0003_00007', 'GEN4.1111.00001.i0001_00006',
                  'GENO.1017.00001.i0002_00006', 'GENO.1017.00001.i0002_00007',
                  'GENO.1216.00002.i0001_00007'],
            '9': ['GEN2.1017.00001.i0004_00012', 'GENO.1017.00001.i0002_00008'],
            '10': ['GEN2.1017.00001.i0003_00008', 'GEN4.1111.00001.i0001_00007',
                   'GENO.1017.00001.i0002_00009', 'GENO.1216.00002.b0001_00008'],
            '11': ['GEN2.1017.00001.i0003_00009', 'GEN4.1111.00001.i0001_00008',
                   'GENO.1017.00001.i0002_00010', 'GENO.1216.00002.b0002_00009'],
            '12': ['GEN2.1017.00001.b0002_00003', 'GENO.1216.00002.i0001_00004'],
            '13': ['GEN2.1017.00001.b0001_00002', 'GEN4.1111.00001.i0001_00004',
                   'GENO.1017.00001.i0002_00004', 'GENO.1216.00002.i0001_00005'],
            '14': ['GEN2.1017.00001.b0001_00001', 'GEN4.1111.00001.i0001_00005',
                   'GENO.1017.00001.i0002_00005', 'GENO.1216.00002.i0001_00006'],
            '15': ['GENO.1216.00002.b0003_00011'],
            '16': ['GENO.1216.00002.b0003_00012']
            }

ALL_STRAINS = ['GEN2.1017.00001', 'GEN4.1111.00001', 'GENO.1017.00001', 'GENO.1216.00002']


def test_read_gene():
    """
    Check that when reading a given gene name, it extracts expected information
    """
    gene = "ESCO.1016.00012.i012_00015"
    num = "1"
    fams_by_strain = {"1": {}}
    all_strains = set()
    upan.read_gene(gene, num, fams_by_strain, all_strains)
    assert all_strains == {"ESCO.1016.00012"}
    assert fams_by_strain == {"1": {"ESCO.1016.00012": [gene]}}


def test_read_gene_other_format():
    """
    Check that when reading a gene name which does not have the gembase
    format, it still returns the right information
    """
    gene = "my_gene-name_other_0001"
    num = "1"
    fams_by_strain = {"1": {}}
    all_strains = set()
    upan.read_gene(gene, num, fams_by_strain, all_strains)
    assert all_strains == {"my_gene-name_other"}
    assert fams_by_strain == {"1": {"my_gene-name_other": [gene]}}


def test_read_gene_strain_known():
    """
    Check that when reading a gene name, and its corresponding
    strain already exists in fams_by_strain and all_strains, it just adds
    this new gene to fams_by_strain, and does nothing to all_strains
    """
    gene = "ESCO.1016.00012.i012_00015"
    num = "1"
    fams_by_strain = {"1": {"ESCO.1016.00012": ["ESCO.1016.00012.i001_01"]}}
    all_strains = {"ESCO.1016.00012"}
    upan.read_gene(gene, num, fams_by_strain, all_strains)
    assert all_strains == {"ESCO.1016.00012"}
    assert fams_by_strain == {"1": {"ESCO.1016.00012": ["ESCO.1016.00012.i001_01",
                                                        gene]}}


def test_read_panfile(caplog):
    """
    check that it reads the pangenome file and returns the expected objects
    """
    caplog.set_level(logging.DEBUG)
    logger = logging.getLogger("test_pan")
    fbs, fams, sas = upan.read_pan_file(PAN_FILE, logger)
    assert fbs == FAMS_BY_STRAIN
    assert fams == FAMILIES
    assert sas == ALL_STRAINS
    assert "Reading and getting information from pangenome file" in caplog.text


def test_get_fams(caplog):
    """
    Test that when giving all members for each family, it returns all strains involved in
    those families, and all families with members sorted by strain
    """
    caplog.set_level(logging.INFO)
    logger = logging.getLogger("test_pan")
    bystrain, sortedf = upan.get_fams_info(FAMILIES, logger)
    assert bystrain == FAMS_BY_STRAIN
    assert sortedf == ALL_STRAINS
    assert "Retrieving information from pan families" in caplog.text


def test_read_pan_filetxt(caplog):
    """
    Test that when giving a pangenome file, it returns all families as expected.
    """
    caplog.set_level(logging.INFO)
    logger = logging.getLogger("test_pan")
    # Copy pan file to folder for files generated by tests. It will also save its bin version
    pan_to_use = os.path.join(GENEPATH, "Pangenome.lst")
    shutil.copyfile(PAN_FILE, pan_to_use)
    fbs, fams, ass = upan.read_pangenome(pan_to_use, logger)
    assert fbs == FAMS_BY_STRAIN
    assert fams == FAMILIES
    assert ass == ALL_STRAINS
    assert "Reading and getting information from pangenome file" in caplog.text
    assert "Saving all information to a binary file for later use" in caplog.text
    assert os.path.isfile(pan_to_use + ".bin")


def test_read_pan_filebin(caplog):
    """
    Test that when giving only a pangenome filename, and the corresponding bin file exists,
    it reads the binary file, and returns expected objects.
    """
    caplog.set_level(logging.INFO)
    logger = logging.getLogger("test_pan")
    # Copy pan file to folder for files generated by tests. It will also save its bin version
    pan_to_use = os.path.join(GENEPATH, "Pangenome.lst")
    panbin_to_use = os.path.join(GENEPATH, "Pangenome.lst.bin")
    test_panbin = os.path.join(PAN_TEST, "pangenome.bin")
    shutil.copyfile(PAN_FILE, pan_to_use)
    shutil.copyfile(test_panbin, panbin_to_use)
    fbs, fams, ass = upan.read_pangenome(pan_to_use, logger)
    assert fbs == FAMS_BY_STRAIN
    assert fams == FAMILIES
    assert ass == ALL_STRAINS
    assert "Retrieving info from binary file" in caplog.text


def test_read_pan_fams(caplog):
    """
    Test that when giving a pangenome file, and families, it directly extracts strain information
    from the families: pangenome file does not need to exist, and a binary file is created
    """
    caplog.set_level(logging.DEBUG)
    logger = logging.getLogger("test_pan")
    panfile = os.path.join(GENEPATH, "toto.txt")
    fbs, fams, ass = upan.read_pangenome(panfile, logger, families=FAMILIES)
    assert fbs == FAMS_BY_STRAIN
    assert fams == FAMILIES
    assert ass == ALL_STRAINS
    assert "Retrieving information from pan families" in caplog.text
    assert "Saving all information to a binary file for later use" in caplog.text
    assert os.path.isfile(panfile + ".bin")
    

def test_read_pan_fams_binok(caplog):
    """
    Test that when giving a pangenome file, and families, it directly extracts strain information
    from the families: pangenome file does not need to exist. However, the pangenome.bin file
    already exists (whatever its content), and is then not recreated.
    """
    caplog.set_level(logging.INFO)
    logger = logging.getLogger("test_pan")
    panfile = os.path.join(GENEPATH, "toto.txt")
    # Create bn pangenome file (which is empty
    open(panfile + ".bin", "w").close()
    fbs, fams, ass = upan.read_pangenome(panfile, logger, FAMILIES)
    assert fbs == FAMS_BY_STRAIN
    assert fams == FAMILIES
    assert ass == ALL_STRAINS
    with open(panfile + ".bin", "r") as panf:
        all_lines = panf.readlines()
    assert all_lines == []
    assert "Retrieving information from pan families" in caplog.text
    assert os.path.isfile(panfile + ".bin")



#!/usr/bin/env python3
# coding: utf-8

# ###############################################################################
# This file is part of PanACOTA.                                                #
#                                                                               #
# Authors: Amandine Perrin                                                      #
# Copyright © 2018-2020 Institut Pasteur (Paris).                               #
# See the COPYRIGHT file for details.                                           #
#                                                                               #
# PanACOTA is a software providing tools for large scale bacterial comparative  #
# genomics. From a set of complete and/or draft genomes, you can:               #
#    -  Do a quality control of your strains, to eliminate poor quality         #
# genomes, which would not give any information for the comparative study       #
#    -  Uniformly annotate all genomes                                          #
#    -  Do a Pan-genome                                                         #
#    -  Do a Core or Persistent genome                                          #
#    -  Align all Core/Persistent families                                      #
#    -  Infer a phylogenetic tree from the Core/Persistent families             #
#                                                                               #
# PanACOTA is free software: you can redistribute it and/or modify it under the #
# terms of the Affero GNU General Public License as published by the Free       #
# Software Foundation, either version 3 of the License, or (at your option)     #
# any later version.                                                            #
#                                                                               #
# PanACOTA is distributed in the hope that it will be useful, but WITHOUT ANY   #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS     #
# FOR A PARTICULAR PURPOSE. See the Affero GNU General Public License           #
# for more details.                                                             #
#                                                                               #
# You should have received a copy of the Affero GNU General Public License      #
# along with PanACOTA (COPYING file).                                           #
# If not, see <https://www.gnu.org/licenses/>.                                  #
# ###############################################################################

"""
Functions to use mmseqs to create a pangenome

@author gem
April 2017
"""
import logging
import os
import time
import multiprocessing
import progressbar
import copy

from PanACoTA import utils

logger = logging.getLogger("pangenome.mmseqs")


def run_all_pangenome(min_id, clust_mode, outdir, prt_path, threads, panfile=None, quiet=False):
    """
    Run all steps to build a pangenome:

    - create mmseqs database from protein bank
    - cluster proteins
    - convert to pangenome


    Parameters
    ----------
    min_id : float
        minimum percentage of identity to be in the same family
    clust_mode : [0, 1, 2]
        0 for 'set cover', 1 for 'single-linkage', 2 for 'CD-Hit'
    outdir : str
        directory where output cluster file must be saved
    prt_path : str
        path to file containing all proteins to cluster.
    threads : int
        number of threads which can be used
    panfile : str or None
        name for output pangenome file. Otherwise, will use default name
    quiet : bool
        True if nothing must be written on stdout, False otherwise.

    Returns
    -------
    (families, outfile) : tuple

        - families : {fam_num: [all members]}
        - outfile : pangenome filename
    """
    # Get general information and file/directory names
    prt_bank = os.path.basename(prt_path)
    start = time.strftime('%Y-%m-%d_%H-%M-%S')
    mmseqdb = os.path.join(outdir, prt_bank + "-msDB")
    information = ("Will run MMseqs2 with:\n"
                   f"\t- minimum sequence identity = {min_id*100}%\n"
                   f"\t- cluster mode {clust_mode}")
    if threads > 1:
        information += "\n\t- {} threads".format(threads)
    logger.info(information)

    infoname = get_info(threads, min_id, clust_mode, start)
    logmmseq = get_logmmseq(outdir, prt_bank, infoname)
    # Create ffindex of DB if not already done
    create_mmseqs_db(mmseqdb, prt_path, logmmseq)

    # Cluster with mmseqs
    if panfile:
        panfile = os.path.join(outdir, panfile)
    families, outfile = do_pangenome(outdir, prt_bank, mmseqdb, min_id,
                                     clust_mode, threads, start, panfile, quiet)
    return families, outfile


def get_info(threads, min_id, clust_mode, start):
    """
    Get string containing all information on future run

    Parameters
    ----------
    threads : int
        max number of threads to use
    min_id : float
        min percentage of identity to consider 2 proteins in hte same family
    clust_mode : [0, 1, 2]
        0 for 'set cover', 1 for 'single-linkage', 2 for 'CD-Hit'
    start : str
        string containing starting date and time

    Returns
    -------
    str
        string containing info on current run, to put in filenames
    """
    if threads != 1:
        threadinfo = "-th" + str(threads)
    else:
        threadinfo = ""
    infoname = str(min_id) + "-mode" + str(clust_mode) + threadinfo + "_" + start
    return infoname


def get_logmmseq(outdir, prt_bank, infoname):
    """
    Get filename of logfile, given information

    Parameters
    ----------
    outdir : str
        output directory
    prt_bank : str
        name of file (without path) containing all proteins to cluster
    infoname : str
        string containing information on the current run

    Returns
    -------
    str
        path to mmseq logfile
    """
    return os.path.join(outdir, "mmseq_" + prt_bank + "_" + infoname + ".log")


def do_pangenome(outdir, prt_bank, mmseqdb, min_id, clust_mode, threads, start, panfile=None,
                 quiet=False):
    """
    Use mmseqs to cluster proteins

    Parameters
    ----------
    outdir : str
        directory where output files are saved
    prt_bank : str
        name of the file containing all proteins to cluster, without path
    mmseqdb : str
        path to base filename for output mmseq db
    min_id : float
        min percentage of identity to be considered in the same family (between 0 and 1)
    clust_mode : [0, 1, 2]
        0 for 'set cover', 1 for 'single-linkage', 2 for 'CD-Hit'
    threads : int
        max number of threads to use
    start : str
        string containing trat date and time
    panfile : str
        if a pangenome file is specified. Otherwise, default pangenome name will be used
    quiet : bool
        true if nothing must be print on stdout/stderr, false otherwise (show progress bar)

    Returns
    -------
    (families, outfile) : tuple

        - families : {fam_num: [all members]}
        - outfile : pangenome filename
    """
    infoname = get_info(threads, min_id, clust_mode, start)
    logmmseq = get_logmmseq(outdir, prt_bank, infoname)
    mmseqclust = os.path.join(outdir, prt_bank + "-clust-" + infoname)
    tmpdir = os.path.join(outdir, "tmp_" + prt_bank + "_" + infoname)
    os.makedirs(tmpdir, exist_ok=True)
    bar = None
    logger.debug(mmseqclust)
    if os.path.isfile(mmseqclust):
        logger.warning(("mmseqs clustering {} already exists. The program will now convert "
                        "it to a pangenome file.").format(mmseqclust))
    else:
        logger.info("Clustering proteins...")
        if not quiet:
            widgets = [progressbar.BouncingBar(marker=progressbar.RotatingMarker(markers="◐◓◑◒")),
                       "  -  ", progressbar.Timer()]
            bar = progressbar.ProgressBar(widgets=widgets, max_value=20, term_width=50)
        pool = multiprocessing.Pool(1)
        args = [mmseqdb, mmseqclust, tmpdir, logmmseq, min_id, threads, clust_mode]
        final = pool.map_async(run_mmseqs_clust, [args], chunksize=1)
        pool.close()
        if not quiet:
            while True:
                if final.ready():
                    break
                bar.update()
            bar.finish()
        pool.join()
    # Convert output to tsv file (one line per comparison done)
    #  # Convert output to tsv file (one line per comparison done)
    # -> returns (families, outfile)
    return mmseqs_to_pangenome(mmseqdb, mmseqclust, logmmseq, start, panfile)


def run_mmseqs_clust(args):
    """
    Run mmseqs clustering

    Parameters
    ----------
    args : tuple
         (mmseqdb, mmseqclust, tmpdir, logmmseq, min_id, threads, clust_mode), with:

            * mmseqdb: path to base filename for output mmseq db
            * mmseqclust: path to base filename for output of mmseq clustering
            * tmpdir : path to folder which will contain mmseq temporary files
            * logmmseq : path to file where logs must be written
            * min_id : min percentage of identity to be considered in the same family
            *         (between 0 and 1)
            * threads : max number of threads to use
            * clust_mode : [0, 1, 2], 0 for 'set cover', 1 for 'single-linkage', 2 for 'CD-Hit'

    """
    mmseqdb, mmseqclust, tmpdir, logmmseq, min_id, threads, clust_mode = args
    cmd = ("mmseqs cluster {} {} {} --min-seq-id {} --threads {} --cluster-mode "
           "{}").format(mmseqdb, mmseqclust, tmpdir, min_id, threads, clust_mode)
    logger.details(f"MMseqs command: {cmd}")
    msg = f"Problem while clustering proteins with mmseqs. See log in {logmmseq}"
    with open(logmmseq, "a") as logm:
        utils.run_cmd(cmd, msg, eof=False, stdout=logm, stderr=logm)


def mmseqs_to_pangenome(mmseqdb, mmseqclust, logmmseq, start, outfile=None):
    """
    Convert mmseqs clustering to a pangenome file:

    - convert mmseqs results to tsv file
    - convert tsv file to pangenome

    Parameters
    ----------
    mmseqdb : str
         path to base filename for output mmseq db
    mmseqclust : str
        path to base filename for output of mmseq clustering
    logmmseq : str
         path to file where logs must be written
    start : str
        string containing start date and time
    outfile : str
        pangenome filename, or None if default one must be used

    Returns
    -------
    (families, outfile) : tuple

        - families : {fam_num: [all members]}
        - outfile : pangenome filename
    """
    cmd = f"mmseqs createtsv {mmseqdb} {mmseqdb} {mmseqclust} {mmseqclust}.tsv"
    msg = "Problem while trying to convert mmseq result file to tsv file"
    logger.details(f"MMseqs command: {cmd}")
    with open(logmmseq, "a") as logf:
        utils.run_cmd(cmd, msg, eof=True, stdout=logf, stderr=logf)
    # Convert the tsv file to a 'pangenome' file: one line per family
    families, outfile = mmseqs_tsv_to_pangenome(mmseqclust, logmmseq, start, outfile)
    return families, outfile


def mmseqs_tsv_to_pangenome(mmseqclust, logmmseq, start, outfile=None):
    """
    Convert the tsv output file of mmseqs to the pangenome file

    Parameters
    ----------
    mmseqclust : str
        path to base filename for output of mmseq clustering
    logmmseq : str
        path to file where logs must be written
    start : str
        string containing start date and time
    outfile : str
        pangenome filename, or None if default one must be used

    Returns
    -------
    (families, outfile) : tuple

        - families : {fam_num: [all members]}
        - outfile : pangenome filename
    """
    logger.info("Converting mmseqs results to pangenome file")
    tsvfile = mmseqclust + ".tsv"
    if not outfile:
        outpath = os.path.dirname(tsvfile)
        base = os.path.basename(tsvfile)
        outfile = os.path.join(outpath, "PanGenome-" + base + ".lst")
    clusters = mmseq_tsv_to_clusters(tsvfile)
    families = clusters_to_file(clusters, outfile)
    end = time.strftime('%Y-%m-%d_%H-%M-%S')
    with open(logmmseq, "a") as logm:
        logm.write(f"\n------------\n\nStart: {start} \n")
        logm.write(f"End: {end}")
    return families, outfile


def mmseq_tsv_to_clusters(mmseq):
    """
    Reads the output of mmseq as a tsv file, and converts it to a python dict

    Parameters
    ----------
    mmseq : str
        filename of mmseq clustering output in tsv format

    Returns
    -------
    dict
        {representative_of_cluster: [list of members]}

    """
    clusters = {}  # {representative: [all members]}
    with open(mmseq) as mmsf:
        for line in mmsf:
            repres, other = line.strip().split()
            if repres in clusters:
                clusters[repres].append(other)
            else:
                clusters[repres] = [repres]
    logger.info("Pangenome has {} families.".format(len(clusters)))
    return clusters


def clusters_to_file(clust, fileout):
    """
    Write all clusters to a file

    Parameters
    ----------
    clust : {first_member: [all members = protein names]}
    fileout : filename of pangenome where families must be written

    Returns
    -------
    dict
        families : {famnum: [members]}
    """
    families = {}  # {famnum: [members]}
    with open(fileout, "w") as fout:
        num = 1
        for _, fam in clust.items():
            families[num] = []
            fout.write(str(num))
            for mem in sorted(fam, key=utils.sort_proteins):
                families[num].append(mem)
                fout.write(" " + mem)
            fout.write("\n")
            num += 1
    return families


def create_mmseqs_db(mmseqdb, prt_path, logmmseq):
    """
    Create ffindex of protein bank (prt_path) if not already done. If done, just write a message
    to tell the user that the current existing file will be used.

    Parameters
    ----------
    mmseqdb : str
         path to base filename for output mmseq db
    prt_path : str
        path to the file containing all proteins to cluster
    logmmseq : str
         path to file where logs must be written
    """
    outext = ["", ".index", ".dbtype", ".lookup", "_h", "_h.index", "_h.dbtype"]
    files_existing = []
    if os.path.isfile(mmseqdb):
        for file in [mmseqdb + ext for ext in outext]:
            if not os.path.isfile(file):
                continue
            files_existing.append(file)
        if len(files_existing) != len(outext):
            logger.warning(f"mmseqs database {mmseqdb} already exists, but at least 1 associated "
                            "file (.dbtype, .index etc). is missing. The program will "
                            "remove existing files and recreate the database.")
            files_remaining = copy.deepcopy(files_existing)
            for file in files_existing:
                os.remove(file)  # Delete file
                files_remaining.remove(file)  # Remove file from list of existing files
                logger.details(f"Removing '{file}'.")
            files_existing = copy.deepcopy(files_remaining)
        else:
            logger.warning(f"mmseqs database {mmseqdb} already exists. The program will "
                           "use it.")
            return 0
    if len(files_existing) != len(outext):
        logger.info("Creating database")
        logger.debug("Existing files: {}".format(len(files_existing)))
        logger.debug("Expected extensions: {}".format(len(outext)))
        cmd = f"mmseqs createdb {prt_path} {mmseqdb}"
        msg = (f"Problem while trying to convert database {prt_path} to mmseqs "
               "database format.")
        logger.details(f"MMseqs command: {cmd}")
        with open(logmmseq, "w") as logf:
            utils.run_cmd(cmd, msg, eof=True, stdout=logf, stderr=logf)



#!/usr/bin/env python3
# coding: utf-8

"""
Unit tests for quicktree_func submodule of tree_module
"""

import os
import pytest
import shutil
import logging

import PanACoTA.tree_module.quicktree_func as qt
from PanACoTA import utils
from . import utilities as tree_util
import test.test_unit.utilities_for_tests as tutil

# Define common variables
ALPATH = os.path.join("test", "data", "align")
ALIGNMENT = os.path.join(ALPATH, "exp_files", "exp_pers4genomes.grp.aln")
TREEPATH = os.path.join("test", "data", "tree")
EXPPATH = os.path.join(TREEPATH, "exp_files")
GENEPATH = os.path.join(TREEPATH, "generated_by_unit-tests")
LOGFILE_BASE = "log_test_fastme"
LOGFILES = [LOGFILE_BASE + ext for ext in [".log", ".log.debug", ".log.details", ".log.err"]]


@pytest.fixture(autouse=True)
def setup_teardown_module():
    """
    Remove log files at the end of this test module

    Before each test:
    - init logger
    - create directory to put generated files

    After:
    - remove all log files
    - remove directory with generated results
    """
    utils.init_logger(LOGFILE_BASE, 0, 'test_quicktree', verbose=1)
    os.mkdir(GENEPATH)
    print("setup")

    yield
    for f in LOGFILES:
        if os.path.exists(f):
            os.remove(f)
    shutil.rmtree(GENEPATH)
    print("teardown")


def test_convert_stockholm(caplog):
    """
    Test that when giving a valid fasta alignment file, it converts it to Stockholm format,
    as expected.
    """
    caplog.set_level(logging.DEBUG)
    outfile = os.path.join(GENEPATH, "test_2stockholm")
    qt.convert2stockholm(ALIGNMENT, outfile)
    exp_stk = os.path.join(EXPPATH, "exp_align_stockholm.stk")
    assert os.path.isfile(outfile)
    tutil.compare_order_content(outfile, exp_stk)
    assert "Converting fasta alignment to stockholm format" in caplog.text


def test_convert_exists(caplog):
    """
    Test that when asking to convert a file in stockholm format, but output file already exists,
    it does not convert again, and writes warning message saying that current file will be used.
    """
    caplog.set_level(logging.DEBUG)
    exp_stk = os.path.join(EXPPATH, "exp_align_stockholm.stk")
    assert qt.convert2stockholm(ALIGNMENT, exp_stk) is None
    assert 'Stockholm alignment file already existing.' in caplog.text
    assert ("The Stockholm alignment file test/data/tree/exp_files/exp_align_stockholm.stk "
            "already exists. The program will use it instead of re-converting "
            "test/data/align/exp_files/exp_pers4genomes.grp.aln") in caplog.text


def test_run_quicktree_default(caplog):
    """
    Test running quicktree without bootstrap
    """
    caplog.set_level(logging.DEBUG)
    align = os.path.join(EXPPATH, "exp_align_stockholm.stk")
    boot = None
    qt.run_quicktree(align, boot, GENEPATH)
    assert "Running Quicktree..." in caplog.text
    assert ("quicktree -in a -out t  "
            "test/data/tree/exp_files/exp_align_stockholm.stk") in caplog.text
    treefile = os.path.join(GENEPATH, "exp_align_stockholm.stk.quicktree_tree.nwk")
    log_file = os.path.join(GENEPATH, "exp_align_stockholm.stk.quicktree.log")
    assert os.path.isfile(log_file)
    assert tree_util.is_tree_lengths(treefile)
    assert not tree_util.is_tree_bootstrap(treefile)


def test_run_quicktree_boot_notree(caplog):
    """
    Test running quicktree with bootstraps, and no given name for treefile
    """
    caplog.set_level(logging.DEBUG)
    align = os.path.join(EXPPATH, "exp_align_stockholm.stk")
    boot = 111
    qt.run_quicktree(align, boot, GENEPATH)
    assert "Running Quicktree..." in caplog.text
    assert ("quicktree -in a -out t -boot 111 "
            "test/data/tree/exp_files/exp_align_stockholm.stk") in caplog.text
    treefile = os.path.join(GENEPATH, "exp_align_stockholm.stk.quicktree_tree.nwk")
    log_file = os.path.join(GENEPATH, "exp_align_stockholm.stk.quicktree.log")
    assert os.path.isfile(log_file)
    assert os.path.isfile(treefile)
    assert not tree_util.is_tree_lengths(treefile)
    assert tree_util.is_tree_bootstrap(treefile)


def test_run_twice(caplog):
    """
    Test generating phylogenetic tree from fasta alignment with quicktree
    Running a second time, it uses the alignment file already generated
    """
    caplog.set_level(logging.DEBUG)
    boot = 111
    qt.run_tree(ALIGNMENT, boot, GENEPATH)
    assert "Converting fasta alignment to stockholm format." in caplog.text
    assert "Running Quicktree..." in caplog.text
    assert ("quicktree -in a -out t -boot 111 "
            "test/data/tree/generated_by_unit-tests/"
            "exp_pers4genomes.grp.aln.stockholm") in caplog.text
    stock = os.path.join(GENEPATH, "exp_pers4genomes.grp.aln.stockholm")
    assert os.path.isfile(stock)
    log_file = os.path.join(GENEPATH, "exp_pers4genomes.grp.aln.stockholm.quicktree.log")
    assert os.path.isfile(log_file)
    treefile = os.path.join(GENEPATH, "exp_pers4genomes.grp.aln.stockholm.quicktree_tree.nwk")
    assert not tree_util.is_tree_lengths(treefile)
    assert tree_util.is_tree_bootstrap(treefile)

    boot = None
    qt.run_tree(ALIGNMENT, boot, GENEPATH)
    assert "Stockholm alignment file already existing." in caplog.text
    assert ("The Stockholm alignment file "
            "test/data/tree/generated_by_unit-tests/exp_pers4genomes.grp.aln.stockholm "
            "already exists. The program will use it "
            "instead of re-converting "
            "test/data/align/exp_files/exp_pers4genomes.grp.aln") in caplog.text
    assert "Running Quicktree..." in caplog.text
    assert ("quicktree -in a -out t  test/data/tree/generated_by_unit-tests/"
            "exp_pers4genomes.grp.aln.stockholm") in caplog.text
    assert os.path.isfile(stock)
    assert os.path.isfile(log_file)
    assert tree_util.is_tree_lengths(treefile)
    assert not tree_util.is_tree_bootstrap(treefile)



#!/usr/bin/env python3
# coding: utf-8

"""
Functional tests for the parser of 'prepare' subcommand
"""
import argparse
import pytest

from PanACoTA.subcommands import prepare


def test_parser_noarg(capsys):
    """
    Test that when the script is called without any argument, an error message appears,
    indicating the required arguments.
    """
    parser = argparse.ArgumentParser(description="Prepare genomes", add_help=False)
    prepare.build_parser(parser)
    with pytest.raises(SystemExit):
        prepare.parse(parser, "".split())
    _, err = capsys.readouterr()
    print(err)
    assert "error: " in err
    assert ("As you did not put the '--norefseq' nor the '-M' option, it means that you want "
            "to download refseq (or genbank) genomes. But you did not provide any information, so PanACoTA "
            "cannot guess which species you want to download. Specify NCBI_taxid (-t)") in err
    assert ("NCBI species taxid (-T) and/or NCBI_species (-g) to download, "
            "or add one of the 2 options (--norefseq or -M) "
            "if you want to skip the 'download step'.") in err


def test_cutn_noint(capsys):
    """
    Test that when user is giving a number of 'N' from which to cut which is:
     - not a number
     - not an int
     - <0
    it gives error message
    """
    parser = argparse.ArgumentParser(description="Prepare genomes", add_help=False)
    prepare.build_parser(parser)
    # Not a number
    with pytest.raises(SystemExit):
        prepare.parse(parser, "--cutn ten".split())
    _, err = capsys.readouterr()
    assert "error: argument --cutn: invalid int value: 'ten'" in err
    # Not an int
    with pytest.raises(SystemExit):
        prepare.parse(parser, "--cutn 1.5".split())
    _, err = capsys.readouterr()
    assert "error: argument --cutn: invalid int value: '1.5'" in err
    # Negative number
    with pytest.raises(SystemExit):
        prepare.parse(parser, "--cutn -5".split())
    _, err = capsys.readouterr()
    assert "error: argument --cutn must be a positive integer: invalid int value: '-5'" in err


def test_l90_noint(capsys):
    """
    Test that when user is giving a number for max L90 which is not valid:
     - not a number
     - not an int
    it gives error message
    """
    parser = argparse.ArgumentParser(description="Prepare genomes", add_help=False)
    prepare.build_parser(parser)
    # Not a number
    with pytest.raises(SystemExit):
        prepare.parse(parser, "--l90 ten".split())
    _, err = capsys.readouterr()
    assert "error: argument --l90: invalid int value: 'ten'" in err
    # Not an int
    with pytest.raises(SystemExit):
        prepare.parse(parser, "--l90 1.5".split())
    _, err = capsys.readouterr()
    assert "error: argument --l90: invalid int value: '1.5'" in err


def test_parser_negative_cont(capsys):
    """
    Test that when the script is called with a limit of contig number <0,
    it returns an error message
    """
    parser = argparse.ArgumentParser(description="Prepare", add_help=False)
    prepare.build_parser(parser)
    with pytest.raises(SystemExit):
        prepare.parse(parser, "--nbcont -5".split())
    _, err = capsys.readouterr()
    assert "The maximum number of contigs allowed must be a positive number." in err


def test_parser_high_cont(capsys):
    """
    Test that when the script is called with a negative limit of contig number,
    it returns an error message
    """
    parser = argparse.ArgumentParser(description="Prepare", add_help=False)
    prepare.build_parser(parser)
    with pytest.raises(SystemExit):
        prepare.parse(parser, "--nbcont 10005".split())
    _, err = capsys.readouterr()
    assert "We do not support genomes with more than 9999 contigs." in err


def test_parser_wrong_cont(capsys):
    """
    Test that when the script is called with a non integer limit of contig number,
    it returns an error message
    """
    parser = argparse.ArgumentParser(description="prepare", add_help=False)
    prepare.build_parser(parser)
    with pytest.raises(SystemExit):
        prepare.parse(parser, "--nbcont 10.5".split())
    _, err = capsys.readouterr()
    assert "argument --nbcont: invalid int value: 10.5" in err


def test_parser_wrong_level(capsys):
    """
    Test that when the script is called with a non integer limit of contig number,
    it returns an error message
    """
    parser = argparse.ArgumentParser(description="prepare", add_help=False)
    prepare.build_parser(parser)
    with pytest.raises(SystemExit):
        prepare.parse(parser, "-t 1234 -o toto -l toto".split())
    _, err = capsys.readouterr()
    assert ("Please choose between available assembly levels: 'all', 'complete', "
            "'chromosome', 'scaffold', 'contig'. If several levels, provide a "
            "comma-separated list. Invalid value: 'toto'") in err


def test_parser_wrong_level_notcomma(capsys):
    """
    Test that when the script is called with a non integer limit of contig number,
    it returns an error message
    """
    parser = argparse.ArgumentParser(description="prepare", add_help=False)
    prepare.build_parser(parser)
    with pytest.raises(SystemExit):
        prepare.parse(parser, "-t 1234 -o outdir -l complete.scaffold".split())
    _, err = capsys.readouterr()
    assert ("Please choose between available assembly levels: 'all', 'complete', "
            "'chromosome', 'scaffold', 'contig'. If several levels, provide a "
            "comma-separated list. Invalid value: 'complete.scaffold'") in err


def test_max_mash_dist(capsys):
    """
    Test that when user is giving a number for max_dist which is not valid:
     - not a number
     - > 1
     - <0
    it gives error message
    """
    parser = argparse.ArgumentParser(description="Prepare genomes", add_help=False)
    prepare.build_parser(parser)
    # Not a number
    with pytest.raises(SystemExit):
        prepare.parse(parser, "--max_dist ten".split())
    _, err = capsys.readouterr()
    assert "error: mash distance: invalid float value: 'ten'" in err
    # > 1
    with pytest.raises(SystemExit):
        prepare.parse(parser, "--max_dist 1.5".split())
    _, err = capsys.readouterr()
    assert "error: mash distance must be between 0 and 1: invalid value: '1.5'" in err
    # < 0
    with pytest.raises(SystemExit):
        prepare.parse(parser, "--max_dist -0.5".split())
    _, err = capsys.readouterr()
    assert "error: mash distance must be between 0 and 1: invalid value: '-0.5'" in err


def test_min_mash_dist(capsys):
    """
    Test that when user is giving a number for max_dist which is not valid:
     - not a number
     - > 1
     - <0
    it gives error message
    """
    parser = argparse.ArgumentParser(description="Prepare genomes", add_help=False)
    prepare.build_parser(parser)
    # Not a number
    with pytest.raises(SystemExit):
        prepare.parse(parser, "--min_dist ten".split())
    _, err = capsys.readouterr()
    assert "error: mash distance: invalid float value: 'ten'" in err
    # > 1
    with pytest.raises(SystemExit):
        prepare.parse(parser, "--min_dist 1.5".split())
    _, err = capsys.readouterr()
    assert "error: mash distance must be between 0 and 1: invalid value: '1.5'" in err
    # < 0
    with pytest.raises(SystemExit):
        prepare.parse(parser, "--min_dist -0.5".split())
    _, err = capsys.readouterr()
    assert "error: mash distance must be between 0 and 1: invalid value: '-0.5'" in err


def test_min_sup_max(capsys):
    '''
    Test that we get an error message if min_dist > max_dist
    '''
    parser = argparse.ArgumentParser(description="Prepare genomes", add_help=False)
    prepare.build_parser(parser)
    # Not a number
    with pytest.raises(SystemExit):
        prepare.parse(parser, "--min_dist 0.9 --max_dist=0.8 --norefseq -o toto".split())
    _, err = capsys.readouterr()
    assert "min_dist (0.9) cannot be higher than max_dist (0.8)" in err


def test_parser_wrong_thread(capsys):
    """
    Test that when the user does not give an int for the threads value, it returns an
    error message.
    """
    parser = argparse.ArgumentParser(description="Prepare", add_help=False)
    prepare.build_parser(parser)
    with pytest.raises(SystemExit):
        prepare.parse(parser, "-p 10.5".split())
    _, err = capsys.readouterr()
    assert "argument --threads threads: invalid int value: 10.5" in err
    # Negative number of threads
    with pytest.raises(SystemExit):
        prepare.parse(parser, "-p -1".split())
    _, err = capsys.readouterr()
    assert ("Please provide a positive number of threads (or 0 for all threads): "
            "Invalid value: -1") in err


def test_parser_more_threads(capsys):
    """
    Test that when the user does not give an int for the threads value, it returns an
    error message.
    """
    import multiprocessing
    nb_cpu = multiprocessing.cpu_count()
    parser = argparse.ArgumentParser(description="Prepare", add_help=False)
    prepare.build_parser(parser)
    with pytest.raises(SystemExit):
        prepare.parse(parser, "-p 30000".split())
    _, err = capsys.readouterr()
    assert (f"You have {nb_cpu} threads on your computer, you cannot ask for more: "
            "invalid value: 30000") in err


def test_parser_all_threads(capsys):
    """
    Test that when the user does not give an int for the threads value, it returns an
    error message.
    """
    import multiprocessing
    nb_cpu = multiprocessing.cpu_count()
    parser = argparse.ArgumentParser(description="Prepare", add_help=False)
    prepare.build_parser(parser)
    options = prepare.parse(parser, "-p 0 --norefseq -o toto".split())
    assert options.parallel == nb_cpu
    assert options.norefseq == True
    assert options.only_mash == False


def test_parse_missing_arg(capsys):
    """
    running prepare without NCBI info nor mash_only nor norefseq -> error asking one of those
    """
    parser = argparse.ArgumentParser(description="Prepare", add_help=False)
    prepare.build_parser(parser)
    with pytest.raises(SystemExit):
        prepare.parse(parser, "-p 1".split())
    _, err = capsys.readouterr()
    assert ("As you did not put the '--norefseq' nor the '-M' option, it means that you want "
            "to download refseq (or genbank) genomes. But you did not provide any information, so PanACoTA "
            "cannot guess which species you want to download. Specify NCBI_taxid (-t)") in err
    assert ("NCBI species taxid (-T) and/or NCBI_species (-g) to download, "
            "or add one of the 2 options (--norefseq or -M) "
            "if you want to skip the 'download step'.") in err


def test_norefseq_nooutdir(capsys):
    """
    Try running without refseq, but not giving an output directory
    """
    parser = argparse.ArgumentParser(description="Prepare", add_help=False)
    prepare.build_parser(parser)
    with pytest.raises(SystemExit):
        prepare.parse(parser, "--norefseq".split())
    _, err = capsys.readouterr()
    assert ("You must provide an output directory, where your results will be saved.") in err


def test_onlymash_noinfo(capsys):
    """
    Try running without refseq, but not giving an output directory
    """
    parser = argparse.ArgumentParser(description="Prepare", add_help=False)
    prepare.build_parser(parser)
    with pytest.raises(SystemExit):
        prepare.parse(parser, "-M".split())
    _, err = capsys.readouterr()
    assert ("If you want to run only Mash filtering steps, please give the info file with "
            "the required information (see '--info' option") in err


def test_onlymash_nooutdir(capsys):
    """
    Try running without refseq, but not giving an output directory
    """
    parser = argparse.ArgumentParser(description="Prepare", add_help=False)
    prepare.build_parser(parser)
    with pytest.raises(SystemExit):
        prepare.parse(parser, "-M --info toto ".split())
    _, err = capsys.readouterr()
    assert ("If you want to run only Mash filtering steps, please give the output "
            "directory where you want to save your results (see '-o' option)") in err


def test_verbose_quiet(capsys):
    """
    Try running without refseq, but not giving an output directory
    """
    parser = argparse.ArgumentParser(description="Prepare", add_help=False)
    prepare.build_parser(parser)
    with pytest.raises(SystemExit):
        prepare.parse(parser, "-q -vv -M --info toto -o outdir".split())
    _, err = capsys.readouterr()
    assert ("Choose between a verbose output (-v) or a quiet output (-q). "
            "You cannot have both.") in err


def test_parser_nospecies(capsys):
    """
    Test that when the user does not give an int for the threads value, it returns an
    error message.
    """
    parser = argparse.ArgumentParser(description="Prepare", add_help=False)
    prepare.build_parser(parser)
    options = prepare.parse(parser, "-T 1234".split())
    assert not options.norefseq
    assert not options.only_mash
    assert options.ncbi_species_taxid == "1234"
    assert options.ncbi_taxid == ""
    assert options.ncbi_species_name == ""
    out, err = capsys.readouterr()
    assert ("WARNING: you did not provide a species name ('-g species' option) "
            "nor an output directory ('-o outdir'). "
            "All files will be downloaded in a folder called with the NCBI species "
            "taxid 1234 instead of the species name.") in out


def test_parser_nospecies_nospeid(capsys):
    """
    Test that when the user does not give an int for the threads value, it returns an
    error message.
    """
    parser = argparse.ArgumentParser(description="Prepare", add_help=False)
    prepare.build_parser(parser)
    options = prepare.parse(parser, "-t 1234".split())
    assert not options.norefseq
    assert not options.only_mash
    assert options.ncbi_species_taxid == ""
    assert options.ncbi_taxid == "1234"
    assert options.ncbi_species_name == ""
    out, err = capsys.readouterr()
    assert ("WARNING: you did not provide a species name ('-g species' option) "
            "nor a species taxid ('-T spetaxid') nor an output directory ('-o outdir'). "
            "All files will be downloaded in a folder called with the NCBI "
            "taxid 1234.") in out


def test_parser_default_cutn(capsys):
    """
    Test that when the user does not give an int for the threads value, it returns an
    error message.
    """
    parser = argparse.ArgumentParser(description="Prepare", add_help=False)
    prepare.build_parser(parser)
    options = prepare.parse(parser, "-t 1234 -o outdir -g species".split())
    assert not options.norefseq
    assert not options.only_mash
    assert options.ncbi_taxid == "1234"
    assert options.ncbi_species_name == "species"
    out, err = capsys.readouterr()
    assert ("!! Your genomes will be split when sequence contains at "
            "least 5'N' in a row. If you want to change this threshold, use "
            "'--cutn n' option (n=0 if you do not want to cut)") in out


def test_parser_error_section_name(capsys):
    """
    Test when we give another value than refseq or genbank to -s section option
    """
    parser = argparse.ArgumentParser(description="Prepare", add_help=False)
    prepare.build_parser(parser)
    with pytest.raises(SystemExit):
        prepare.parse(parser, "-t 1234 -o outdir -s species --cutn 1".split())
    _, err = capsys.readouterr()
    assert ("argument -s: invalid choice: 'species' (choose from 'refseq', 'genbank')") in err


def test_parser_default_l90_nb_cont(capsys):
    """
    Test that when the user does not give an int for the threads value, it returns an
    error message.
    """
    parser = argparse.ArgumentParser(description="Prepare", add_help=False)
    prepare.build_parser(parser)
    options = prepare.parse(parser, "-t 1234 -o outdir -g species --cutn 1".split())
    assert not options.norefseq
    assert not options.only_mash
    assert options.ncbi_taxid == "1234"
    assert options.ncbi_species_name == "species"
    out, err = capsys.readouterr()
    assert ("!! Your genomes will be filtered, and only the ones with 'L90' <= 100 "
            "and 'number of contigs' < 999 will be kept. If you want to change those "
            "thresholds, use '--l90' and '--nbcont' options.") in out


def test_parser_info_notonlymash(capsys):
    """
    Giving an info file, but not asking for only_mash -> useless info file
    """
    parser = argparse.ArgumentParser(description="Prepare", add_help=False)
    prepare.build_parser(parser)
    options = prepare.parse(parser, "-T 1234 -o outdir -g species --cutn 1 --info toto".split())
    assert not options.norefseq
    assert not options.only_mash
    assert options.ncbi_species_taxid == "1234"
    assert options.ncbi_species_name == "species"
    out, err = capsys.readouterr()
    assert ("!! You gave an info file (--info option), but did not ask to run only Mash "
            "step (-M option). Your info file will be ignored (and renamed with '.back' "
            "at the end), and another one will be created with the new calculated values.") in out



#!/usr/bin/env python3
# coding: utf-8

"""
Unit tests for the download_genomes_func submodule in prepare module
"""
import os
import logging
import glob
import shutil
import pytest

import PanACoTA.prepare_module.download_genomes_func as downg


DATA_TEST_DIR = os.path.join("test", "data", "prepare")
GENEPATH = os.path.join(DATA_TEST_DIR, "generated_by_unit-tests")


@pytest.fixture(autouse=True)
def setup_teardown_module():
    """
    Remove log files at the end of this test module

    Before each test:
    - init logger
    - create directory to put generated files

    After:
    - remove all log files
    - remove directory with generated results
    """
    if os.path.isdir(GENEPATH):
        content = os.listdir(GENEPATH)
        for f in content:
            assert f.startswith(".fuse")
    else:
        os.mkdir(GENEPATH)
    print("setup")

    yield
    shutil.rmtree(GENEPATH, ignore_errors=True)
    print("teardown")


def test_to_database():
    """
    Test that all fna.gz files are uncompressed and moved to a created Database_init folder
    """
    out_dir = os.path.join(DATA_TEST_DIR, "genomes")
    nb_gen, db_init_dir = downg.to_database(out_dir, "refseq")
    db_dir = os.path.join(DATA_TEST_DIR, "genomes", "Database_init")
    assert os.path.isdir(db_dir)
    files_all = glob.glob(os.path.join(db_dir, "*"))
    files_fna = glob.glob(os.path.join(db_dir, "*.fna"))
    # Check that there are only 3 files in result database
    assert len(files_all) == len(files_fna)
    # And that those files are .fna files
    assert len(files_fna) == 3
    # Check that we have as many genomes as expected, and that the output database has the
    # expected name
    assert nb_gen == 3
    assert db_init_dir == db_dir
    assert os.path.isfile(os.path.join(db_dir, "ACOR002.0519.fna"))
    assert os.path.isfile(os.path.join(db_dir, "ACOR002.0519.fna"))
    assert os.path.isfile(os.path.join(db_dir, "ACOR003.0519.fna"))

    shutil.rmtree(db_dir)


def test_to_database_nofolder_refseq(caplog):
    """
    Test behavior when the folder that should contain refseq downloaded genomes does not exist
    -> should exit with error message
    """
    caplog.set_level(logging.DEBUG)
    with pytest.raises(SystemExit):
        downg.to_database(GENEPATH, "genbank")

    assert "ERROR" in caplog.text
    assert ("The folder containing genomes downloaded from NCBI genbank "
            "(test/data/prepare/generated_by_unit-tests/genbank/bacteria) "
            "does not exist.") in caplog.text
    assert ("Check that you really downloaded sequences (fna.gz) and that they are "
            "in this folder") in caplog.text


def test_to_database_nofolder_per_genome(caplog):
    """
    Test behavior when the folder refseq/bacteria exists, but there are no folders inside
    -> should exit with error message
    """
    outdir = os.path.join(GENEPATH, "out-refseq")
    empty_dir = os.path.join(outdir, "refseq", "bacteria")
    os.makedirs(empty_dir)
    caplog.set_level(logging.DEBUG)
    with pytest.raises(SystemExit):
        downg.to_database(outdir, "refseq")
    # Check error message is as expected
    assert "ERROR" in caplog.text
    assert ("The folder supposed to contain genomes downloaded from NCBI refseq "
            "(test/data/prepare/generated_by_unit-tests/out-refseq/refseq/bacteria) "
            "exists but is empty") in caplog.text
    assert ("Check that you really downloaded sequences (fna.gz)") in caplog.text

    # Same with genbank
    outdir = os.path.join(GENEPATH, "out-genbank")
    empty_dir = os.path.join(outdir, "genbank", "bacteria")
    os.makedirs(empty_dir)
    caplog.set_level(logging.DEBUG)
    with pytest.raises(SystemExit):
        downg.to_database(outdir, "genbank")
    # Check error message is as expected
    assert "ERROR" in caplog.text
    assert ("The folder supposed to contain genomes downloaded from NCBI genbank "
            "(test/data/prepare/generated_by_unit-tests/out-genbank/genbank/bacteria) "
            "exists but is empty") in caplog.text
    assert ("Check that you really downloaded sequences (fna.gz)") in caplog.text


def test_to_database_1empty_genome_folder(caplog):
    """
    Test behavior when the folder refseq/bacteria exists, there are subfolders inside,
    but 1 of them is empty: warning message informing that this genome will be ignored
    """
    caplog.set_level(logging.DEBUG)
    out_dir = os.path.join(GENEPATH, "1empty_genome_folder")
    refseq_dir = os.path.join(DATA_TEST_DIR, "genomes")
    # Copy content of refseq in genomes test data to output folder that will be used
    shutil.copytree(refseq_dir, out_dir)

    # Empty 1 directory: move its file to 'out_dir'
    to_remove = os.path.join(out_dir, "refseq", "bacteria", "ACOR003", "ACOR003.0519.fna.gz")
    os.remove(to_remove)
    # Run to_database
    nb_gen, db_dir = downg.to_database(out_dir, "refseq")
    assert nb_gen == 2
    assert db_dir == os.path.join(out_dir, "Database_init")

    # Check that a warning message was raised, indicating that genome is ignored
    assert "WARNING" in caplog.text
    assert ("Problem with genome in ACOR003: no compressed fasta file downloaded. "
            "This genome will be ignored.") in caplog.text
    assert not os.path.isfile(os.path.join(db_dir, "ACOR003.0519.fna"))
    assert os.path.isfile(os.path.join(db_dir, "ACOR001.0519.fna"))
    assert os.path.isfile(os.path.join(db_dir, "ACOR002.0519.fna"))


def test_to_database_several_genomes(caplog):
    """
    Test behavior when the folder refseq/bacteria exists, there are subfolders inside,
    but 1 of them contains more than 1 genome: warning message informing that this
    genome will be ignored
    """
    out_dir = os.path.join(GENEPATH, "genomes")
    refseq_dir = os.path.join(DATA_TEST_DIR, "genomes")
    # Copy content of refseq in genomes test data to output folder that will be used
    shutil.copytree(refseq_dir, out_dir)

    # Create a new gz file in one of the genome directories
    to_create_filename = "ACOR002.0519.bis.fna.gz" # Name of file that must be created
    to_fill_dir = "ACOR002" # Directory containing file to create
    to_create_path = os.path.join(out_dir, "refseq", "bacteria", to_fill_dir, to_create_filename)
    # Create empty gz file
    open(to_create_path, "w").close()

    # Run to_database, and check that only 2 genomes were considered
    nb_gen, db_dir = downg.to_database(out_dir, "refseq")
    assert nb_gen == 2
    assert db_dir == os.path.join(out_dir, "Database_init")

    # Check that a warning message was raised, indicating that genome is ignored
    caplog.set_level(logging.DEBUG)
    assert "WARNING" in caplog.text
    assert ("Problem with genome in ACOR002: several compressed fasta files found. "
            "This genome will be ignored.") in caplog.text
    assert not os.path.isfile(os.path.join(db_dir, "ACOR002.0519.fna"))
    assert os.path.isfile(os.path.join(db_dir, "ACOR001.0519.fna"))
    assert os.path.isfile(os.path.join(db_dir, "ACOR003.0519.fna"))


def test_to_database_1genome_wrong_format(caplog):
    """
    Test behavior when the folder refseq/bacteria exists, there is 1 genome per subfolder,
    but 1 genome cannot be unzipped
    """
    # out_dir = os.path.join(DATA_TEST_DIR, "genomes")
    # gz_genomes_folder = os.path.join(out_dir, "refseq", "bacteria")

    out_dir = os.path.join(GENEPATH, "genomes")
    refseq_dir = os.path.join(DATA_TEST_DIR, "genomes")
    # Copy content of refseq in genomes test data to output folder that will be used
    shutil.copytree(refseq_dir, out_dir)

    # Name of directory directly containing the original gz file
    to_corrupt_dir = "ACOR001"
    to_corrupt_filename = "ACOR001.0519.fna.gz"
    to_corrupt_path = os.path.join(out_dir, "refseq", "bacteria",  to_corrupt_dir,
                                   to_corrupt_filename)
    # Create fake gz file (txt file)
    false_gz = open(to_corrupt_path, "w")
    false_gz.write("This is not a gz file")
    false_gz.close()

    # Run to_database
    nb_gen, db_dir = downg.to_database(out_dir, "refseq")
    assert nb_gen == 2
    assert db_dir == os.path.join(out_dir, "Database_init")

    # Check that a error message was raised, indicating that genome is ignored
    caplog.set_level(logging.DEBUG)
    assert "ERROR" in caplog.text
    assert ("Error while trying to uncompress "
            "test/data/prepare/generated_by_unit-tests/genomes/Database_init/ACOR001.0519.fna.gz. "
            "This genome will be ignored") in caplog.text
    # Check that there are only 2 files in the database, and that they correspond
    # to uncompressed gz files
    list_db = os.listdir(db_dir)
    assert len(list_db) == 2
    assert not os.path.isfile(os.path.join(db_dir, to_corrupt_filename))
    assert os.path.isfile(os.path.join(db_dir, "ACOR002.0519.fna"))
    assert os.path.isfile(os.path.join(db_dir, "ACOR003.0519.fna"))


def test_download_specify_level(caplog):
    """
    Test that, given a taxid, and a species name,
    it downloads genomes in .gz, and uncompress them in the
    db folder (which is named as expected)

    We cannot compare log, as it is already catched by NCBI_genome_download
    """
    caplog.set_level(logging.INFO)

    species_linked = "Acetobacter_orleanensis"
    section = "refseq"
    NCBI_species = "Acetobacter orleanensis"
    NCBI_species_taxid = "104099"
    NCBI_taxid = ""
    outdir = os.path.join(GENEPATH, "test_download_refseq")
    threads = 1
    levels = ""

    db_dir, nb_gen = downg.download_from_ncbi(species_linked, section, NCBI_species, NCBI_species_taxid, NCBI_taxid, levels,
                                                outdir, threads)
    # Check path to uncompressed files is as expected
    assert db_dir == os.path.join(outdir, "Database_init")
    # Check number of genomes downloaded. We cannot know the exact value, as it is updated
    # everyday. But in nov. 2019, there are 4 genomes. So, there must be at least those 4 genomes
    assert nb_gen >= 4
    # And that db_dir exists and contains nb_gen files
    assert os.path.isdir(db_dir)
    assert len(os.listdir(db_dir)) == nb_gen

    # Check that assembly summary file wwas donwloaded as expected
    sum_file = os.path.join(outdir, "assembly_summary-Acetobacter_orleanensis.txt" )
    assert os.path.isfile(sum_file)
    # Check number of genomes in summary file, and how many with scaffold or complete
    # assembly level -> will check that when asking only for those levels, we get the same number
    other = 0
    scaf = 0
    comp = 0
    with open(sum_file, "r") as sf:
        sf.readline()  # skip header
        for line in sf:
            if "complete" in line.split("\t")[13].lower():
                comp += 1
            elif "scaffold" in line.split("\t")[13].lower():
                scaf += 1
            else:
                other += 1
    assert other + scaf + comp == nb_gen

    # Check that the NCBI_genome_download output directory exists
    ngd_outdir = os.path.join(outdir, "refseq", "bacteria")
    # And that it contains folders
    assert os.path.isdir(ngd_outdir)
    assert len(os.listdir(ngd_outdir)) >= 4
    # Check log giving species name + species taxid
    assert 'Downloading all genomes for NCBI species = Acetobacter orleanensis (NCBI_species_taxid = 104099)' in caplog.text

    # Re-run, but only asking for complete and scaffold
    outdir2 = os.path.join(GENEPATH, "test_download_refseq_only-scaf")
    levels2 = "scaffold,complete"
    db_dir2, nb_gen2 = downg.download_from_ncbi(species_linked, section, NCBI_species, NCBI_species_taxid, NCBI_taxid,
                                                  levels2, outdir2, threads)
    assert scaf + comp == nb_gen2
    assert db_dir2 == os.path.join(outdir2, "Database_init")
    # Check log giving species name + species taxid + levels given
    assert ("Downloading all genomes for NCBI species = Acetobacter orleanensis "
            "(NCBI_species_taxid = 104099). (Only those assembly levels: scaffold,complete)") in caplog.text


def test_download_only_spetaxid(caplog):
    """
    Test that, given a species taxid, it downloads all genomes of the species in .gz, and uncompress them in the
    db folder (which is named as expected)

    We cannot compare log, as it is already catched by NCBI_genome_download
    """
    caplog.set_level(logging.INFO)
    species_linked = "toto"
    section = "refseq"
    NCBI_species = None
    NCBI_species_taxid = "104099"
    NCBI_taxid = ""
    outdir = os.path.join(GENEPATH, "test_download_refseq_noSpe")
    threads = 1
    levels = ""

    db_dir, nb_gen = downg.download_from_ncbi(species_linked, section, NCBI_species, NCBI_species_taxid, NCBI_taxid, levels,
                                                outdir, threads)

    # Check path to uncompressed files is as expected
    assert db_dir == os.path.join(outdir, "Database_init")
    # Check number of genomes downloaded. We cannot know the exact value, as it is updated everyday. But in nov. 2019, there are 4 genomes. So, there must be at least those 4 genomes
    assert nb_gen >= 4
    # And that db_dir exists and contains nb_gen files
    assert os.path.isdir(db_dir)
    assert len(os.listdir(db_dir)) == nb_gen
    # Check log giving only species taxid
    assert "Downloading all genomes for  NCBI_species_taxid = 104099" in caplog.text

    # Check that assembly summary file was donwloaded as expected
    sum_file = os.path.join(outdir, "assembly_summary-toto.txt" )
    assert os.path.isfile(sum_file)

    # Check that the NCBI_genome_download output directory exists
    ngd_outdir = os.path.join(outdir, "refseq", "bacteria")
    # And that it contains folders
    assert os.path.isdir(ngd_outdir)
    assert len(os.listdir(ngd_outdir)) >= 4


def test_download_taxid_and_spetaxid(caplog):
    """
    Test that, given a species taxid and a taxid, it downloads only the genome(s) corresponding to taxid (intersection)

    We cannot compare log, as it is already catched by NCBI_genome_download
    """
    caplog.set_level(logging.INFO)
    species_linked = "toto-spe"
    section = "refseq"
    NCBI_species = None
    NCBI_species_taxid = "104099"
    NCBI_taxid = "1231342"
    levels = ""
    threads = 1
    outdir2 = os.path.join(GENEPATH, "test_download_refseq_noSpeandSpecific")
    db_dir2, nb_gen2 = downg.download_from_ncbi(species_linked, section, NCBI_species, NCBI_species_taxid, NCBI_taxid, levels,
                                                  outdir2, threads)

    # Check path to uncompressed files is as expected
    assert db_dir2 == os.path.join(outdir2, "Database_init")
    # Check number of genomes downloaded. We cannot know the exact value, as it is updated everyday. But in nov. 2019, there are 4 genomes. So, there must be at least those 4 genomes
    assert nb_gen2 == 1
    # And that db_dir exists and contains nb_gen files
    assert os.path.isdir(db_dir2)
    assert len(os.listdir(db_dir2)) == 1
    # Check log giving only species taxid
    assert "Downloading all genomes for  NCBI_species_taxid = 104099 (and NCBI_taxid = 1231342)" in caplog.text

    # Check that assembly summary file was donwloaded as expected
    sum_file = os.path.join(outdir2, "assembly_summary-toto-spe.txt" )
    assert os.path.isfile(sum_file)

    # Check that the NCBI_genome_download output directory exists
    ngd_outdir2 = os.path.join(outdir2, "refseq", "bacteria")
    # And that it contains folders
    assert os.path.isdir(ngd_outdir2)
    assert len(os.listdir(ngd_outdir2)) == 1


def test_download_taxid_and_spename(caplog):
    """
    Test that, given a taxid and a species name, it downloads only the genome(s) corresponding to taxid (intersection)

    We cannot compare log, as it is already catched by NCBI_genome_download
    """
    caplog.set_level(logging.INFO)
    species_linked = "aceor"
    section = "refseq"
    NCBI_species = "Acetobacter orleanensis"
    NCBI_species_taxid = ""
    NCBI_taxid = "1231342"
    levels = ""
    threads = 1
    outdir = os.path.join(GENEPATH, "test_download_refseq_noSpeandSpecific")
    db_dir, nb_gen = downg.download_from_ncbi(species_linked, section, NCBI_species, NCBI_species_taxid, NCBI_taxid, levels,
                                                outdir, threads)

    # Check path to uncompressed files is as expected
    assert db_dir == os.path.join(outdir, "Database_init")
    # Check number of genomes downloaded. We cannot know the exact value, as it is updated everyday. But in nov. 2019, there are 4 genomes. So, there must be at least those 4 genomes
    assert nb_gen == 1
    # And that db_dir exists and contains nb_gen files
    assert os.path.isdir(db_dir)
    assert len(os.listdir(db_dir)) == 1
    # Check log giving only species taxid
    assert "Downloading all genomes for NCBI species = Acetobacter orleanensis (and NCBI_taxid = 1231342)" in caplog.text

    # Check that assembly summary file was donwloaded as expected
    sum_file = os.path.join(outdir, "assembly_summary-aceor.txt" )
    assert os.path.isfile(sum_file)

    # Check that the NCBI_genome_download output directory exists
    ngd_outdir = os.path.join(outdir, "refseq", "bacteria")
    # And that it contains folders
    assert os.path.isdir(ngd_outdir)
    assert len(os.listdir(ngd_outdir)) == 1


def test_download_specific_strain(caplog):
    """
    Test that, given a taxid of a specific strain, it only downloads this one

    We cannot compare log, as it is already catched by NCBI_genome_download
    """
    caplog.set_level(logging.INFO)
    species_linked = "toto"
    NCBI_species = None
    section = "refseq"
    NCBI_species_taxid = ""
    NCBI_taxid = "1123862"
    outdir = os.path.join(GENEPATH, "test_download_refseq_specific")
    threads = 1
    levels = ""

    db_dir, nb_gen = downg.download_from_ncbi(species_linked, section, NCBI_species, NCBI_species_taxid, NCBI_taxid, levels,
                                                outdir, threads)

    # Check path to uncompressed files is as expected
    assert db_dir == os.path.join(outdir, "Database_init")
    # Check number of genomes downloaded. We cannot know the exact value, as it is updated everyday. But in nov. 2019, there are 4 genomes. So, there must be at least those 4 genomes
    assert nb_gen == 1
    # And that db_dir exists and contains nb_gen files
    assert os.path.isdir(db_dir)
    assert len(os.listdir(db_dir)) == nb_gen

    # Check that assembly summary file was donwloaded as expected
    sum_file = os.path.join(outdir, "assembly_summary-toto.txt" )
    assert os.path.isfile(sum_file)

    # Check that the NCBI_genome_download output directory exists
    ngd_outdir = os.path.join(outdir, "refseq", "bacteria")
    # And that it contains folders
    assert os.path.isdir(ngd_outdir)
    assert len(os.listdir(ngd_outdir)) == 1

    # Check log giving only specific taxid
    assert "Downloading all genomes for  NCBI_taxid = 1123862" in caplog.text


def test_download_2taxid(caplog):
    """
    Give a taxid of a subspecies and a taxid of a specific strain. Should download all genomes
    of the subspecies + specific strain.
    If only the subspecies taxid, the specific strain is not downloaded.
    """
    caplog.set_level(logging.INFO)
    species_linked = "salmo"
    section = "refseq"
    NCBI_species = None
    NCBI_species_taxid = ""
    # 913079 is the subspecies Salmonella enterica subsp. enterica serovar Mississippi
    # 1212561  is the strain Salmonella enterica subsp. enterica serovar Mississippi strain 2010K-1406
    NCBI_taxid = "913079,1212561"
    outdir = os.path.join(GENEPATH, "test_download_refseq_2taxid")
    threads = 1
    levels = ""
    db_dir, nb_gen = downg.download_from_ncbi(species_linked, section, NCBI_species, NCBI_species_taxid, NCBI_taxid, levels,
                                                outdir, threads)

    # Check path to uncompressed files is as expected
    assert db_dir == os.path.join(outdir, "Database_init")
    # Check number of genomes downloaded. We cannot know the exact value, as it is updated everyday. But in nov. 2019, there are 4 genomes. So, there must be at least those 4 genomes
    assert nb_gen >= 13
    # And that db_dir exists and contains nb_gen files
    assert os.path.isdir(db_dir)
    assert len(os.listdir(db_dir)) == nb_gen
    # Check log giving only species taxid
    assert "Downloading all genomes for  NCBI_taxid = 913079,1212561" in caplog.text

    # Check that assembly summary file was donwloaded as expected
    sum_file = os.path.join(outdir, "assembly_summary-salmo.txt" )
    assert os.path.isfile(sum_file)

    # Check that the NCBI_genome_download output directory exists
    ngd_outdir = os.path.join(outdir, "refseq", "bacteria")
    assert os.path.isdir(ngd_outdir)


    # Redo, without the specific strain taxid. Should download the same -1 (not the specific strain)
    species_linked = "salmo"
    section = "refseq"
    NCBI_species = None
    NCBI_species_taxid = ""
    # 913079 is the subspecies Salmonella enterica subsp. enterica serovar Mississippi
    NCBI_taxid_1 = "913079"
    outdir_1 = os.path.join(GENEPATH, "test_download_refseq_2taxid_1")
    threads = 1
    levels = ""
    db_dir_1, nb_gen_1 = downg.download_from_ncbi(species_linked, section, NCBI_species, NCBI_species_taxid, NCBI_taxid_1, levels,
                                                    outdir_1, threads)
    assert nb_gen == nb_gen_1 + 1
    assert "Downloading all genomes for  NCBI_taxid = 913079" in caplog.text


def test_download_refseq_vs_genbank(caplog):
    """
    Give a taxid of a subspecies download strains from refseq, and then from genbank.
    Currently, no strains in refseq, and 2 in genbank.
    39831 = Klebsiella pneumoniae subsp. rhinoscleromatis
    Later, there can be some in refseq, but always at least 2 more in genbank
    """
    caplog.set_level(logging.INFO)
    species_linked = "refseq-genbank"
    section = "refseq"
    NCBI_species = None
    NCBI_species_taxid = ""
    NCBI_taxid = "39831"
    outdir = os.path.join(GENEPATH, "test_download_refseq_genbank")
    levels = ""
    threads = 1

    # With refseq, no genome found
    with pytest.raises(SystemExit):
        downg.download_from_ncbi(species_linked, section, NCBI_species, NCBI_species_taxid, NCBI_taxid, levels,
                                   outdir, threads)

    # Check path to uncompressed files does not exist
    assert not os.path.isdir(os.path.join(outdir, "Database_init"))

    # Check that the NCBI_genome_download output directory was not created
    ngd_outdir = os.path.join(outdir, "refseq", "bacteria")
    assert not os.path.isdir(ngd_outdir)

    # Check logs
    assert "ERROR" in caplog.text
    assert ("Could not download genomes. Check that you gave valid NCBI taxid and/or "
            "NCBI species name. If you gave both, check that given taxID and name really "
            "correspond to the same species.") in caplog.text

    # REDO with genbank instead of refseq
    section = "genbank"
    outdir2 = os.path.join(GENEPATH, "test_download_genbank")
    db_dir, nb_gen = downg.download_from_ncbi(species_linked, section, NCBI_species, NCBI_species_taxid, NCBI_taxid, levels,
                                                outdir2, threads)

    # Check path to uncompressed files is as expected
    assert db_dir == os.path.join(outdir2, "Database_init")
    # Check number of genomes downloaded. We cannot know the exact value, as it is updated everyday. But in nov. 2019, there are 4 genomes. So, there must be at least those 4 genomes
    assert nb_gen >= 2
    # And that db_dir exists and contains nb_gen files
    assert os.path.isdir(db_dir)
    assert len(os.listdir(db_dir)) == nb_gen
    # Check log giving only species taxid
    assert "Downloading all genomes for  NCBI_taxid = 39831" in caplog.text
    # Check that assembly summary file was donwloaded as expected
    sum_file = os.path.join(outdir2, "assembly_summary-refseq-genbank.txt" )
    assert os.path.isfile(sum_file)
    # Check that the NCBI_genome_download output directory exists
    ngd_outdir = os.path.join(outdir2, "genbank", "bacteria")
    assert os.path.isdir(ngd_outdir)


def test_download_wrongTaxID(caplog):
    """
    Test that, when a non existing taxid is given, it exits (with error message)

    We cannot compare log, as it is already catched by NCBI_genome_download
    """
    species_linked = "Acetobacter_orleanensis"
    NCBI_species = None
    section = "refseq"
    NCBI_species_taxid = "10409"
    NCBI_taxid = ""
    outdir = os.path.join(GENEPATH, "test_download_refseq_wrongTaxID")
    threads = 1
    levels = ""
    with pytest.raises(SystemExit):
        downg.download_from_ncbi(species_linked, section, NCBI_species, NCBI_species_taxid, NCBI_taxid, levels,
                                   outdir, threads)

    # Check path to uncompressed files does not exist
    assert not os.path.isdir(os.path.join(outdir, "Database_init"))

    # Check that the NCBI_genome_download output directory was not created
    ngd_outdir = os.path.join(outdir, "refseq", "bacteria")
    assert not os.path.isdir(ngd_outdir)

    # Check logs
    caplog.set_level(logging.DEBUG)
    assert "ERROR" in caplog.text
    assert ("Could not download genomes. Check that you gave valid NCBI taxid and/or "
            "NCBI species name. If you gave both, check that given taxID and name really "
            "correspond to the same species.") in caplog.text

    # Check that output directory was not created
    assert not os.path.isdir(outdir)


def test_download_diffSpeTaxID(caplog):
    """
    Test that, when a taxID and a species name are given, but those 2 elements do not
    match with the same genomes, it exists with error message

    We cannot compare log, as it is already catched by NCBI_genome_download
    """
    species_linked = "Acetobacter_orleanensis"
    section = "refseq"
    NCBI_species = "Acetobacter fabarum"
    NCBI_species_taxid = "104099"
    NCBI_taxid = ""
    outdir = os.path.join(GENEPATH, "test_download_refseq_wrongTaxID")
    threads = 1
    levels = ""
    with pytest.raises(SystemExit):
        downg.download_from_ncbi(species_linked, section, NCBI_species, NCBI_species_taxid, NCBI_taxid, levels,
                                   outdir, threads)

    # Check path to uncompressed files does not exist
    assert not os.path.isdir(os.path.join(outdir, "Database_init"))

    # Check that the NCBI_genome_download output directory was not created
    ngd_outdir = os.path.join(outdir, "refseq", "bacteria")
    assert not os.path.isdir(ngd_outdir)

    # Check logs
    caplog.set_level(logging.DEBUG)
    assert "ERROR" in caplog.text
    assert ("Could not download genomes. Check that you gave valid NCBI taxid and/or "
            "NCBI species name. If you gave both, check that given taxID and name really "
            "correspond to the same species.") in caplog.text

    # Check that output directory was not created
    assert not os.path.isdir(outdir)



#!/usr/bin/env python3
# coding: utf-8

"""
Functional tests for the parser of 'pangenome' subcommand
"""

import pytest
import argparse

from PanACoTA.subcommands import pangenome


def test_parser_noarg(capsys):
    """
    Test that when the script is called without any argument, an error message appears,
    indicating the required arguments.
    """
    parser = argparse.ArgumentParser(description="Do pangenome", add_help=False)
    pangenome.build_parser(parser)
    with pytest.raises(SystemExit):
        pangenome.parse(parser, "".split())
    _, err = capsys.readouterr()
    assert "usage: " in err
    assert "-l LSTINFO_FILE -n DATASET_NAME -d DBPATH" in err
    assert "[-i MIN_ID]" in err
    assert " -o OUTDIR" in err
    assert "[-f OUTFILE] [-c {0,1,2}]" in err
    assert "[-s SPEDIR] [--threads THREADS] [-v]" in err
    assert "[-q] [-h]" in err
    assert "the following arguments are required: -l, -n, -d, -o" in err


def test_perc_id_nonum(capsys):
    """
    Test that when the given percentage of identity is not a number, it returns the
    expected error.
    """
    parser = argparse.ArgumentParser(description="Do pangenome", add_help=False)
    pangenome.build_parser(parser)
    with pytest.raises(SystemExit):
        pangenome.parse(parser, "-l lstinfo -n TEST4 -d dbpath -i toto -o outdir".split())
    _, err = capsys.readouterr()
    assert "argument -i percentage_id: invalid float value: toto" in err


def test_neg_perc_id(capsys):
    """
    Test that when the given percentage of identity is a negative number, it returns the
    expected error.
    """
    parser = argparse.ArgumentParser(description="Do pangenome", add_help=False)
    pangenome.build_parser(parser)
    with pytest.raises(SystemExit):
        pangenome.parse(parser, "-l lstinfo -n TEST4 -d dbpath -i -0.2 -o outdir".split())
    _, err = capsys.readouterr()
    assert "The minimum %% of identity must be in [0, 1]. Invalid value: -0.2" in err


def test_big_perc_id(capsys):
    """
    Test that when the given percentage of identity is greater than 1, it returns the
    expected error.
    """
    parser = argparse.ArgumentParser(description="Do pangenome", add_help=False)
    pangenome.build_parser(parser)
    with pytest.raises(SystemExit):
        pangenome.parse(parser, "-l lstinfo -n TEST4 -d dbpath -i 2 -o outdir".split())
    _, err = capsys.readouterr()
    assert "The minimum %% of identity must be in [0, 1]. Invalid value: 2" in err


def test_wrong_clust_mode(capsys):
    """
    Test that when the given cluster mode does not exist, it returns the expected error message
    """
    parser = argparse.ArgumentParser(description="Do pangenome", add_help=False)
    pangenome.build_parser(parser)
    with pytest.raises(SystemExit):
        pangenome.parse(parser, "-l lstinfo -n TEST4 -d dbpath -i 0.8 -o od -c 3".split())
    _, err = capsys.readouterr()
    assert "argument -c: invalid choice: 3 (choose from 0, 1, 2)" in err


def test_thread_no_int(capsys):
    """
    Test that when the given number of threads is not an int, it returns the expected
    error message.
    """
    parser = argparse.ArgumentParser(description="Do pangenome", add_help=False)
    pangenome.build_parser(parser)
    with pytest.raises(SystemExit):
        pangenome.parse(parser, "-l lstinfo -n TEST4 -d dbpath -o od --threads 1.5".split())
    _, err = capsys.readouterr()
    assert "argument --threads threads: invalid int value: 1.5" in err


def test_thread_too_many(capsys):
    """
    Test that when the given number of threads is greater than the number of available threads,
    it returns the expected error message.
    """
    import multiprocessing
    nb = multiprocessing.cpu_count()
    parser = argparse.ArgumentParser(description="Do pangenome", add_help=False)
    pangenome.build_parser(parser)
    with pytest.raises(SystemExit):
        pangenome.parse(parser, f"-l lstinfo -n TEST4 -d dbpath -o od --threads {nb*10}".split())
    _, err = capsys.readouterr()
    assert ("You have {} threads on your computer, you cannot ask for more: "
            "invalid value: {}").format(nb, nb*10) in err


def test_thread_neg(capsys):
    """
    Test that when the given number of threads is negative,
    it returns the expected error message.
    """
    parser = argparse.ArgumentParser(description="Do pangenome", add_help=False)
    pangenome.build_parser(parser)
    with pytest.raises(SystemExit):
        pangenome.parse(parser, "-l lstinfo -n TEST4 -d dbpath -o od --threads -1".split())
    _, err = capsys.readouterr()
    assert "Please provide a positive number of threads (or 0 for all threads)" in err


def test_parser_default():
    """
    Test that when run with the minimum required arguments, all default values are
    as expected.
    """
    parser = argparse.ArgumentParser(description="Do pangenome", add_help=False)
    pangenome.build_parser(parser)
    options = pangenome.parse(parser, "-l lstinfo -n TEST4 -d dbpath -o od".split())
    assert options.lstinfo_file == "lstinfo"
    assert options.dataset_name == "TEST4"
    assert options.dbpath == "dbpath"
    assert options.min_id == 0.8
    assert options.outdir == "od"
    assert options.clust_mode == 1
    assert not options.spedir
    assert options.threads == 1
    assert not options.outfile
    assert options.verbose == 0
    assert not options.quiet


def test_parser_all_threads():
    """
    Test that when run with the minimum required arguments, + '0' for the thread number,
    all default values are as expected, and the number of threads is equal to the total
    number of threads.
    """
    import multiprocessing
    nb = multiprocessing.cpu_count()
    parser = argparse.ArgumentParser(description="Do pangenome", add_help=False)
    pangenome.build_parser(parser)
    options = pangenome.parse(parser, "-l lstinfo -n TEST4 -d dbpath -o od --threads 0".split())
    assert options.lstinfo_file == "lstinfo"
    assert options.dataset_name == "TEST4"
    assert options.dbpath == "dbpath"
    assert options.min_id == 0.8
    assert options.outdir == "od"
    assert options.clust_mode == 1
    assert not options.spedir
    assert options.threads == nb
    assert not options.outfile
    assert options.verbose == 0
    assert not options.quiet


def test_parser_1thread():
    """
    Test that when run with the minimum required arguments, + '0' for the thread number,
    all default values are as expected, and the number of threads is equal to the total
    number of threads.
    """
    parser = argparse.ArgumentParser(description="Do pangenome", add_help=False)
    pangenome.build_parser(parser)
    options = pangenome.parse(parser, "-l lstinfo -n TEST4 -d dbpath -o od --threads 1".split())
    assert options.lstinfo_file == "lstinfo"
    assert options.dataset_name == "TEST4"
    assert options.dbpath == "dbpath"
    assert options.min_id == 0.8
    assert options.outdir == "od"
    assert options.clust_mode == 1
    assert not options.spedir
    assert options.threads == 1
    assert not options.outfile
    assert options.verbose == 0
    assert not options.quiet
